# Copyright (C) 2009-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tools for dealing with the Launchpad API."""

from __future__ import absolute_import

# Importing this module will be expensive, since it imports launchpadlib and
# its dependencies. However, our plan is to only load this module when it is
# needed by a command that uses it.


import re
try:
    from urllib.parse import (
        urlparse,
        urlunparse,
        )
except ImportError:  # python < 3
    from urlparse import (
        urlparse,
        urlunparse,
        )

from ... import (
    branch,
    bedding,
    errors,
    osutils,
    trace,
    transport,
    )
from ...i18n import gettext


class LaunchpadlibMissing(errors.DependencyNotPresent):

    _fmt = ("launchpadlib is required for Launchpad API access. "
            "Please install the launchpadlib package.")

    def __init__(self, e):
        super(LaunchpadlibMissing, self).__init__(
            'launchpadlib', e)

try:
    import launchpadlib
except ImportError as e:
    raise LaunchpadlibMissing(e)

from launchpadlib.launchpad import (
    Launchpad,
    )
from launchpadlib import uris

# Declare the minimum version of launchpadlib that we need in order to work.
MINIMUM_LAUNCHPADLIB_VERSION = (1, 6, 3)


def get_cache_directory():
    """Return the directory to cache launchpadlib objects in."""
    return osutils.pathjoin(bedding.cache_dir(), 'launchpad')


def parse_launchpadlib_version(version_number):
    """Parse a version number of the style used by launchpadlib."""
    return tuple(map(int, version_number.split('.')))


def check_launchpadlib_compatibility():
    """Raise an error if launchpadlib has the wrong version number."""
    installed_version = parse_launchpadlib_version(launchpadlib.__version__)
    if installed_version < MINIMUM_LAUNCHPADLIB_VERSION:
        raise errors.DependencyNotPresent(
            'launchpadlib',
            'At least launchpadlib %s is required, but installed version is %s'
            % (MINIMUM_LAUNCHPADLIB_VERSION, installed_version))


def lookup_service_root(service_root):
    try:
        return uris.lookup_service_root(service_root)
    except ValueError:
        if service_root != 'qastaging':
            raise
        staging_root = uris.lookup_service_root('staging')
        return staging_root.replace('staging', 'qastaging')


class NoLaunchpadBranch(errors.BzrError):
    _fmt = 'No launchpad branch could be found for branch "%(url)s".'

    def __init__(self, branch):
        errors.BzrError.__init__(self, branch=branch, url=branch.base)


def connect_launchpad(base_url, timeout=None, proxy_info=None,
                      version=Launchpad.DEFAULT_VERSION):
    """Log in to the Launchpad API.

    :return: The root `Launchpad` object from launchpadlib.
    """
    if proxy_info is None:
        import httplib2
        proxy_info = httplib2.proxy_info_from_environment('https')
    try:
        cache_directory = get_cache_directory()
    except EnvironmentError:
        cache_directory = None
    return Launchpad.login_with(
        'breezy', base_url, cache_directory, timeout=timeout,
        proxy_info=proxy_info, version=version)



class LaunchpadBranch(object):
    """Provide bzr and lp API access to a Launchpad branch."""

    def __init__(self, lp_branch, bzr_url, bzr_branch=None, check_update=True):
        """Constructor.

        :param lp_branch: The Launchpad branch.
        :param bzr_url: The URL of the Bazaar branch.
        :param bzr_branch: An instance of the Bazaar branch.
        """
        self.bzr_url = bzr_url
        self._bzr = bzr_branch
        self._push_bzr = None
        self._check_update = check_update
        self.lp = lp_branch

    @property
    def bzr(self):
        """Return the bzr branch for this branch."""
        if self._bzr is None:
            self._bzr = branch.Branch.open(self.bzr_url)
        return self._bzr

    @property
    def push_bzr(self):
        """Return the push branch for this branch."""
        if self._push_bzr is None:
            self._push_bzr = branch.Branch.open(self.lp.bzr_identity)
        return self._push_bzr

    @staticmethod
    def plausible_launchpad_url(url):
        """Is 'url' something that could conceivably be pushed to LP?

        :param url: A URL that may refer to a Launchpad branch.
        :return: A boolean.
        """
        if url is None:
            return False
        if url.startswith('lp:'):
            return True
        regex = re.compile('([a-z]*\\+)*(bzr\\+ssh|http)'
                           '://bazaar.*.launchpad.net')
        return bool(regex.match(url))

    @staticmethod
    def candidate_urls(bzr_branch):
        """Iterate through related URLs that might be Launchpad URLs.

        :param bzr_branch: A Bazaar branch to find URLs from.
        :return: a generator of URL strings.
        """
        url = bzr_branch.get_public_branch()
        if url is not None:
            yield url
        url = bzr_branch.get_push_location()
        if url is not None:
            yield url
        url = bzr_branch.get_parent()
        if url is not None:
            yield url
        yield bzr_branch.base

    @staticmethod
    def tweak_url(url, launchpad):
        """Adjust a URL to work with staging, if needed."""
        if str(launchpad._root_uri) == uris.STAGING_SERVICE_ROOT:
            return url.replace('bazaar.launchpad.net',
                               'bazaar.staging.launchpad.net')
        elif str(launchpad._root_uri) == lookup_service_root('qastaging'):
            return url.replace('bazaar.launchpad.net',
                               'bazaar.qastaging.launchpad.net')
        return url

    @classmethod
    def from_bzr(cls, launchpad, bzr_branch, create_missing=True):
        """Find a Launchpad branch from a bzr branch."""
        check_update = True
        for url in cls.candidate_urls(bzr_branch):
            url = cls.tweak_url(url, launchpad)
            if not cls.plausible_launchpad_url(url):
                continue
            lp_branch = launchpad.branches.getByUrl(url=url)
            if lp_branch is not None:
                break
        else:
            if not create_missing:
                raise NoLaunchpadBranch(bzr_branch)
            lp_branch = cls.create_now(launchpad, bzr_branch)
            check_update = False
        return cls(lp_branch, bzr_branch.base, bzr_branch, check_update)

    @classmethod
    def create_now(cls, launchpad, bzr_branch):
        """Create a Bazaar branch on Launchpad for the supplied branch."""
        url = cls.tweak_url(bzr_branch.get_push_location(), launchpad)
        if not cls.plausible_launchpad_url(url):
            raise errors.BzrError(gettext('%s is not registered on Launchpad') %
                                  bzr_branch.base)
        bzr_branch.create_clone_on_transport(transport.get_transport(url))
        lp_branch = launchpad.branches.getByUrl(url=url)
        if lp_branch is None:
            raise errors.BzrError(gettext('%s is not registered on Launchpad') %
                                  url)
        return lp_branch

    def get_target(self):
        """Return the 'LaunchpadBranch' for the target of this one."""
        lp_branch = self.lp
        if lp_branch.project is not None:
            dev_focus = lp_branch.project.development_focus
            if dev_focus is None:
                raise errors.BzrError(gettext('%s has no development focus.') %
                                      lp_branch.bzr_identity)
            target = dev_focus.branch
            if target is None:
                raise errors.BzrError(gettext(
                    'development focus %s has no branch.') % dev_focus)
        elif lp_branch.sourcepackage is not None:
            target = lp_branch.sourcepackage.getBranch(pocket="Release")
            if target is None:
                raise errors.BzrError(gettext(
                                      'source package %s has no branch.') %
                                      lp_branch.sourcepackage)
        else:
            raise errors.BzrError(gettext(
                '%s has no associated product or source package.') %
                lp_branch.bzr_identity)
        return LaunchpadBranch(target, target.bzr_identity)

    def update_lp(self):
        """Update the Launchpad copy of this branch."""
        if not self._check_update:
            return
        with self.bzr.lock_read():
            if self.lp.last_scanned_id is not None:
                if self.bzr.last_revision() == self.lp.last_scanned_id:
                    trace.note(gettext('%s is already up-to-date.') %
                               self.lp.bzr_identity)
                    return
                graph = self.bzr.repository.get_graph()
                if not graph.is_ancestor(osutils.safe_utf8(self.lp.last_scanned_id),
                                         self.bzr.last_revision()):
                    raise errors.DivergedBranches(self.bzr, self.push_bzr)
                trace.note(gettext('Pushing to %s') % self.lp.bzr_identity)
            self.bzr.push(self.push_bzr)

    def find_lca_tree(self, other):
        """Find the revision tree for the LCA of this branch and other.

        :param other: Another LaunchpadBranch
        :return: The RevisionTree of the LCA of this branch and other.
        """
        graph = self.bzr.repository.get_graph(other.bzr.repository)
        lca = graph.find_unique_lca(self.bzr.last_revision(),
                                    other.bzr.last_revision())
        return self.bzr.repository.revision_tree(lca)


def canonical_url(object):
    """Return the canonical URL for a branch."""
    scheme, netloc, path, params, query, fragment = urlparse(
        str(object.self_link))
    path = '/'.join(path.split('/')[2:])
    netloc = netloc.replace('api.', 'code.')
    return urlunparse((scheme, netloc, path, params, query, fragment))



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# The normalize function is taken from pygettext which is distributed
# with Python under the Python License, which is GPL compatible.

"""Extract docstrings from Bazaar commands.

This module only handles breezy objects that use strings not directly wrapped
by a gettext() call. To generate a complete translation template file, this
output needs to be combined with that of xgettext or a similar command for
extracting those strings, as is done in the bzr Makefile. Sorting the output
is also left to that stage of the process.
"""

from __future__ import absolute_import

import inspect
import os
import sys

import breezy
from . import (
    commands as _mod_commands,
    errors,
    help_topics,
    option,
    plugin as _mod_plugin,
    )
from .sixish import PY3
from .trace import (
    mutter,
    note,
    )
from .i18n import gettext


def _escape(s):
    s = (s.replace('\\', '\\\\')
         .replace('\n', '\\n')
         .replace('\r', '\\r')
         .replace('\t', '\\t')
         .replace('"', '\\"')
         )
    return s


def _normalize(s):
    # This converts the various Python string types into a format that
    # is appropriate for .po files, namely much closer to C style.
    lines = s.split('\n')
    if len(lines) == 1:
        s = '"' + _escape(s) + '"'
    else:
        if not lines[-1]:
            del lines[-1]
            lines[-1] = lines[-1] + '\n'
        lineterm = '\\n"\n"'
        s = '""\n"' + lineterm.join(map(_escape, lines)) + '"'
    return s


def _parse_source(source_text, filename='<unknown>'):
    """Get object to lineno mappings from given source_text"""
    import ast
    cls_to_lineno = {}
    str_to_lineno = {}
    for node in ast.walk(ast.parse(source_text, filename)):
        # TODO: worry about duplicates?
        if isinstance(node, ast.ClassDef):
            # TODO: worry about nesting?
            cls_to_lineno[node.name] = node.lineno
        elif isinstance(node, ast.Str):
            # Python AST gives location of string literal as the line the
            # string terminates on. It's more useful to have the line the
            # string begins on. Unfortunately, counting back newlines is
            # only an approximation as the AST is ignorant of escaping.
            str_to_lineno[node.s] = node.lineno - (0 if sys.version_info >= (3, 8) else node.s.count('\n'))
    return cls_to_lineno, str_to_lineno


class _ModuleContext(object):
    """Record of the location within a source tree"""

    def __init__(self, path, lineno=1, _source_info=None):
        self.path = path
        self.lineno = lineno
        if _source_info is not None:
            self._cls_to_lineno, self._str_to_lineno = _source_info

    @classmethod
    def from_module(cls, module):
        """Get new context from module object and parse source for linenos"""
        sourcepath = inspect.getsourcefile(module)
        # TODO: fix this to do the right thing rather than rely on cwd
        relpath = os.path.relpath(sourcepath)
        return cls(relpath,
                   _source_info=_parse_source("".join(inspect.findsource(module)[0]), module.__file__))

    def from_class(self, cls):
        """Get new context with same details but lineno of class in source"""
        try:
            lineno = self._cls_to_lineno[cls.__name__]
        except (AttributeError, KeyError):
            mutter("Definition of %r not found in %r", cls, self.path)
            return self
        return self.__class__(self.path, lineno,
                              (self._cls_to_lineno, self._str_to_lineno))

    def from_string(self, string):
        """Get new context with same details but lineno of string in source"""
        try:
            lineno = self._str_to_lineno[string]
        except (AttributeError, KeyError):
            mutter("String %r not found in %r", string[:20], self.path)
            return self
        return self.__class__(self.path, lineno,
                              (self._cls_to_lineno, self._str_to_lineno))


class _PotExporter(object):
    """Write message details to output stream in .pot file format"""

    def __init__(self, outf, include_duplicates=False):
        self.outf = outf
        if include_duplicates:
            self._msgids = None
        else:
            self._msgids = set()
        self._module_contexts = {}

    def poentry(self, path, lineno, s, comment=None):
        if self._msgids is not None:
            if s in self._msgids:
                return
            self._msgids.add(s)
        if comment is None:
            comment = ''
        else:
            comment = "# %s\n" % comment
        mutter("Exporting msg %r at line %d in %r", s[:20], lineno, path)
        line = (
            "#: {path}:{lineno}\n"
            "{comment}"
            "msgid {msg}\n"
            "msgstr \"\"\n"
            "\n".format(
                path=path, lineno=lineno, comment=comment, msg=_normalize(s)))
        if not PY3:
            line = line.decode('utf-8')
        self.outf.write(line)

    def poentry_in_context(self, context, string, comment=None):
        context = context.from_string(string)
        self.poentry(context.path, context.lineno, string, comment)

    def poentry_per_paragraph(self, path, lineno, msgid, include=None):
        # TODO: How to split long help?
        paragraphs = msgid.split('\n\n')
        if include is not None:
            paragraphs = filter(include, paragraphs)
        for p in paragraphs:
            self.poentry(path, lineno, p)
            lineno += p.count('\n') + 2

    def get_context(self, obj):
        module = inspect.getmodule(obj)
        try:
            context = self._module_contexts[module.__name__]
        except KeyError:
            context = _ModuleContext.from_module(module)
            self._module_contexts[module.__name__] = context
        if inspect.isclass(obj):
            context = context.from_class(obj)
        return context


def _write_option(exporter, context, opt, note):
    if getattr(opt, 'hidden', False):
        return
    optname = opt.name
    if getattr(opt, 'title', None):
        exporter.poentry_in_context(context, opt.title,
                                    "title of {name!r} {what}".format(name=optname, what=note))
    for name, _, _, helptxt in opt.iter_switches():
        if name != optname:
            if opt.is_hidden(name):
                continue
            name = "=".join([optname, name])
        if helptxt:
            exporter.poentry_in_context(context, helptxt,
                                        "help of {name!r} {what}".format(name=name, what=note))


def _standard_options(exporter):
    OPTIONS = option.Option.OPTIONS
    context = exporter.get_context(option)
    for name in sorted(OPTIONS):
        opt = OPTIONS[name]
        _write_option(exporter, context.from_string(name), opt, "option")


def _command_options(exporter, context, cmd):
    note = "option of {0!r} command".format(cmd.name())
    for opt in cmd.takes_options:
        # String values in Command option lists are for global options
        if not isinstance(opt, str):
            _write_option(exporter, context, opt, note)


def _write_command_help(exporter, cmd):
    context = exporter.get_context(cmd.__class__)
    rawdoc = cmd.__doc__
    dcontext = context.from_string(rawdoc)
    doc = inspect.cleandoc(rawdoc)

    def exclude_usage(p):
        # ':Usage:' has special meaning in help topics.
        # This is usage example of command and should not be translated.
        if p.splitlines()[0] != ':Usage:':
            return True

    exporter.poentry_per_paragraph(dcontext.path, dcontext.lineno, doc,
                                   exclude_usage)
    _command_options(exporter, context, cmd)


def _command_helps(exporter, plugin_name=None):
    """Extract docstrings from path.

    This respects the Bazaar cmdtable/table convention and will
    only extract docstrings from functions mentioned in these tables.
    """

    # builtin commands
    for cmd_name in _mod_commands.builtin_command_names():
        command = _mod_commands.get_cmd_object(cmd_name, False)
        if command.hidden:
            continue
        if plugin_name is not None:
            # only export builtins if we are not exporting plugin commands
            continue
        note(gettext("Exporting messages from builtin command: %s"), cmd_name)
        _write_command_help(exporter, command)

    plugins = _mod_plugin.plugins()
    if plugin_name is not None and plugin_name not in plugins:
        raise errors.BzrError(gettext('Plugin %s is not loaded' % plugin_name))
    core_plugins = set(
        name for name in plugins
        if plugins[name].path().startswith(breezy.__path__[0]))
    # plugins
    for cmd_name in _mod_commands.plugin_command_names():
        command = _mod_commands.get_cmd_object(cmd_name, False)
        if command.hidden:
            continue
        if plugin_name is not None and command.plugin_name() != plugin_name:
            # if we are exporting plugin commands, skip plugins we have not
            # specified.
            continue
        if plugin_name is None and command.plugin_name() not in core_plugins:
            # skip non-core plugins
            # TODO: Support extracting from third party plugins.
            continue
        note(gettext("Exporting messages from plugin command: {0} in {1}").format(
             cmd_name, command.plugin_name()))
        _write_command_help(exporter, command)


def _error_messages(exporter):
    """Extract fmt string from breezy.errors."""
    context = exporter.get_context(errors)
    base_klass = errors.BzrError
    for name in dir(errors):
        klass = getattr(errors, name)
        if not inspect.isclass(klass):
            continue
        if not issubclass(klass, base_klass):
            continue
        if klass is base_klass:
            continue
        if klass.internal_error:
            continue
        fmt = getattr(klass, "_fmt", None)
        if fmt:
            note(gettext("Exporting message from error: %s"), name)
            exporter.poentry_in_context(context, fmt)


def _help_topics(exporter):
    topic_registry = help_topics.topic_registry
    for key in topic_registry.keys():
        doc = topic_registry.get(key)
        if isinstance(doc, str):
            exporter.poentry_per_paragraph(
                'dummy/help_topics/' + key + '/detail.txt',
                1, doc)
        elif callable(doc):  # help topics from files
            exporter.poentry_per_paragraph(
                'en/help_topics/' + key + '.txt',
                1, doc(key))
        summary = topic_registry.get_summary(key)
        if summary is not None:
            exporter.poentry('dummy/help_topics/' + key + '/summary.txt',
                             1, summary)


def export_pot(outf, plugin=None, include_duplicates=False):
    exporter = _PotExporter(outf, include_duplicates)
    if plugin is None:
        _standard_options(exporter)
        _command_helps(exporter)
        _error_messages(exporter)
        _help_topics(exporter)
    else:
        _command_helps(exporter, plugin)



# Copyright (C) 2019 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from ..propose import (
    determine_title,
    get_proposal_by_url,
    get_hoster,
    Hoster,
    MergeProposal,
    UnsupportedHoster,
    )
from .. import (
    propose as _mod_propose,
    registry,
    urlutils,
    )

from .. import (
    tests,
    )


class SampleMergeProposal(MergeProposal):
    """Sample merge proposal."""


class SampleHoster(Hoster):

    _locations = []

    @classmethod
    def _add_location(cls, url):
        cls._locations.append(url)

    @classmethod
    def probe_from_url(cls, url, possible_transports=None):
        for b in cls._locations:
            if url.startswith(b):
                return cls()
        raise UnsupportedHoster(url)

    def hosts(self, branch):
        for b in self._locations:
            if branch.user_url.startswith(b):
                return True
        return False

    @classmethod
    def iter_instances(cls):
        return iter([cls()])

    def get_proposal_by_url(self, url):
        for b in self._locations:
            if url.startswith(b):
                return MergeProposal()
        raise UnsupportedHoster(url)


class SampleHosterTestCase(tests.TestCaseWithTransport):

    def setUp(self):
        super(SampleHosterTestCase, self).setUp()
        self._old_hosters = _mod_propose.hosters
        _mod_propose.hosters = registry.Registry()
        self.hoster = SampleHoster()
        os.mkdir('hosted')
        SampleHoster._add_location(
            urlutils.local_path_to_url(os.path.join(self.test_dir, 'hosted')))
        _mod_propose.hosters.register('sample', self.hoster)

    def tearDown(self):
        super(SampleHosterTestCase, self).tearDown()
        _mod_propose.hosters = self._old_hosters
        SampleHoster._locations = []


class TestGetHosterTests(SampleHosterTestCase):

    def test_get_hoster(self):
        tree = self.make_branch_and_tree('hosted/branch')
        self.assertIs(self.hoster, get_hoster(tree.branch, [self.hoster]))
        self.assertIsInstance(get_hoster(tree.branch), SampleHoster)

        tree = self.make_branch_and_tree('blah')
        self.assertRaises(UnsupportedHoster, get_hoster, tree.branch)


class TestGetProposal(SampleHosterTestCase):

    def test_get_proposal_by_url(self):
        self.assertRaises(UnsupportedHoster, get_proposal_by_url, 'blah')

        url = urlutils.local_path_to_url(os.path.join(self.test_dir, 'hosted', 'proposal'))
        self.assertIsInstance(get_proposal_by_url(url), MergeProposal)


class DetermineTitleTests(tests.TestCase):

    def test_determine_title(self):
        self.assertEqual('Make some change', determine_title("""\
Make some change.

And here are some more details.
"""))



# Copyright (C) 2008-2018 Jelmer Vernooij <jelmer@jelmer.uk>
# Copyright (C) 2007 Canonical Ltd
# Copyright (C) 2008 John Carr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Converters, etc for going between Bazaar and Git ids."""

from __future__ import absolute_import

import base64
import stat

from .. import (
    bencode,
    errors,
    foreign,
    trace,
    urlutils,
    )
from ..foreign import (
    ForeignVcs,
    VcsMappingRegistry,
    ForeignRevision,
    )
from ..revision import (
    NULL_REVISION,
    Revision,
    )
from ..sixish import (
    PY3,
    text_type,
    viewitems,
    )
from .errors import (
    NoPushSupport,
    )
from .hg import (
    format_hg_metadata,
    extract_hg_metadata,
    )
from .roundtrip import (
    extract_bzr_metadata,
    inject_bzr_metadata,
    CommitSupplement,
    )


DEFAULT_FILE_MODE = stat.S_IFREG | 0o644
HG_RENAME_SOURCE = b"HG:rename-source"
HG_EXTRA = b"HG:extra"

# This HG extra is used to indicate the commit that this commit was based on.
HG_EXTRA_AMEND_SOURCE = b"amend_source"

FILE_ID_PREFIX = b'git:'

# Always the same.
ROOT_ID = b"TREE_ROOT"


class UnknownCommitExtra(errors.BzrError):
    _fmt = "Unknown extra fields in %(object)r: %(fields)r."

    def __init__(self, object, fields):
        errors.BzrError.__init__(self)
        self.object = object
        self.fields = ",".join(fields)


class UnknownMercurialCommitExtra(errors.BzrError):
    _fmt = "Unknown mercurial extra fields in %(object)r: %(fields)r."

    def __init__(self, object, fields):
        errors.BzrError.__init__(self)
        self.object = object
        self.fields = b",".join(fields)


def escape_file_id(file_id):
    file_id = file_id.replace(b'_', b'__')
    file_id = file_id.replace(b' ', b'_s')
    file_id = file_id.replace(b'\x0c', b'_c')
    return file_id


def unescape_file_id(file_id):
    ret = bytearray()
    i = 0
    while i < len(file_id):
        if file_id[i:i + 1] != b'_':
            ret.append(file_id[i])
        else:
            if file_id[i + 1:i + 2] == b'_':
                ret.append(b"_"[0])
            elif file_id[i + 1:i + 2] == b's':
                ret.append(b" "[0])
            elif file_id[i + 1:i + 2] == b'c':
                ret.append(b"\x0c"[0])
            else:
                raise ValueError("unknown escape character %s" %
                                 file_id[i + 1:i + 2])
            i += 1
        i += 1
    return bytes(ret)


def fix_person_identifier(text):
    if b"<" not in text and b">" not in text:
        username = text
        email = text
    else:
        if text.rindex(b">") < text.rindex(b"<"):
            raise ValueError(text)
        username, email = text.split(b"<", 2)[-2:]
        email = email.split(b">", 1)[0]
        if username.endswith(b" "):
            username = username[:-1]
    return b"%s <%s>" % (username, email)


def warn_escaped(commit, num_escaped):
    trace.warning("Escaped %d XML-invalid characters in %s. Will be unable "
                  "to regenerate the SHA map.", num_escaped, commit)


def warn_unusual_mode(commit, path, mode):
    trace.mutter("Unusual file mode %o for %s in %s. Storing as revision "
                 "property. ", mode, path, commit)


class BzrGitMapping(foreign.VcsMapping):
    """Class that maps between Git and Bazaar semantics."""
    experimental = False

    BZR_DUMMY_FILE = None

    def is_special_file(self, filename):
        return (filename in (self.BZR_DUMMY_FILE, ))

    def __init__(self):
        super(BzrGitMapping, self).__init__(foreign_vcs_git)

    def __eq__(self, other):
        return (type(self) == type(other)
                and self.revid_prefix == other.revid_prefix)

    @classmethod
    def revision_id_foreign_to_bzr(cls, git_rev_id):
        """Convert a git revision id handle to a Bazaar revision id."""
        from dulwich.protocol import ZERO_SHA
        if git_rev_id == ZERO_SHA:
            return NULL_REVISION
        return b"%s:%s" % (cls.revid_prefix, git_rev_id)

    @classmethod
    def revision_id_bzr_to_foreign(cls, bzr_rev_id):
        """Convert a Bazaar revision id to a git revision id handle."""
        if not bzr_rev_id.startswith(b"%s:" % cls.revid_prefix):
            raise errors.InvalidRevisionId(bzr_rev_id, cls)
        return bzr_rev_id[len(cls.revid_prefix) + 1:], cls()

    def generate_file_id(self, path):
        # Git paths are just bytestrings
        # We must just hope they are valid UTF-8..
        if isinstance(path, text_type):
            path = path.encode("utf-8")
        if path == b"":
            return ROOT_ID
        return FILE_ID_PREFIX + escape_file_id(path)

    def parse_file_id(self, file_id):
        if file_id == ROOT_ID:
            return u""
        if not file_id.startswith(FILE_ID_PREFIX):
            raise ValueError
        return unescape_file_id(file_id[len(FILE_ID_PREFIX):]).decode('utf-8')

    def revid_as_refname(self, revid):
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        if PY3:
            revid = revid.decode('utf-8')
        quoted_revid = urlutils.quote(revid)
        return b"refs/bzr/" + quoted_revid.encode('utf-8')

    def import_unusual_file_modes(self, rev, unusual_file_modes):
        if unusual_file_modes:
            ret = [(path, unusual_file_modes[path])
                   for path in sorted(unusual_file_modes.keys())]
            rev.properties[u'file-modes'] = bencode.bencode(ret)

    def export_unusual_file_modes(self, rev):
        try:
            file_modes = rev.properties[u'file-modes']
        except KeyError:
            return {}
        else:
            return dict(bencode.bdecode(file_modes.encode("utf-8")))

    def _generate_git_svn_metadata(self, rev, encoding):
        try:
            git_svn_id = rev.properties[u"git-svn-id"]
        except KeyError:
            return ""
        else:
            return "\ngit-svn-id: %s\n" % git_svn_id.encode(encoding)

    def _generate_hg_message_tail(self, rev):
        extra = {}
        renames = []
        branch = 'default'
        for name in rev.properties:
            if name == u'hg:extra:branch':
                branch = rev.properties[u'hg:extra:branch']
            elif name.startswith(u'hg:extra'):
                extra[name[len(u'hg:extra:'):]] = base64.b64decode(
                    rev.properties[name])
            elif name == u'hg:renames':
                renames = bencode.bdecode(base64.b64decode(
                    rev.properties[u'hg:renames']))
            # TODO: Export other properties as 'bzr:' extras?
        ret = format_hg_metadata(renames, branch, extra)
        if not isinstance(ret, bytes):
            raise TypeError(ret)
        return ret

    def _extract_git_svn_metadata(self, rev, message):
        lines = message.split("\n")
        if not (lines[-1] == "" and len(lines) >= 2 and
                lines[-2].startswith("git-svn-id:")):
            return message
        git_svn_id = lines[-2].split(": ", 1)[1]
        rev.properties[u'git-svn-id'] = git_svn_id
        (url, rev, uuid) = parse_git_svn_id(git_svn_id)
        # FIXME: Convert this to converted-from property somehow..
        return "\n".join(lines[:-2])

    def _extract_hg_metadata(self, rev, message):
        (message, renames, branch, extra) = extract_hg_metadata(message)
        if branch is not None:
            rev.properties[u'hg:extra:branch'] = branch
        for name, value in viewitems(extra):
            rev.properties[u'hg:extra:' + name] = base64.b64encode(value)
        if renames:
            rev.properties[u'hg:renames'] = base64.b64encode(bencode.bencode(
                [(new, old) for (old, new) in viewitems(renames)]))
        return message

    def _extract_bzr_metadata(self, rev, message):
        (message, metadata) = extract_bzr_metadata(message)
        return message, metadata

    def _decode_commit_message(self, rev, message, encoding):
        return message.decode(encoding), CommitSupplement()

    def _encode_commit_message(self, rev, message, encoding):
        return message.encode(encoding)

    def export_commit(self, rev, tree_sha, parent_lookup, lossy,
                      verifiers):
        """Turn a Bazaar revision in to a Git commit

        :param tree_sha: Tree sha for the commit
        :param parent_lookup: Function for looking up the GIT sha equiv of a
            bzr revision
        :param lossy: Whether to store roundtripping information.
        :param verifiers: Verifiers info
        :return dulwich.objects.Commit represent the revision:
        """
        from dulwich.objects import Commit, Tag
        commit = Commit()
        commit.tree = tree_sha
        if not lossy:
            metadata = CommitSupplement()
            metadata.verifiers = verifiers
        else:
            metadata = None
        parents = []
        for p in rev.parent_ids:
            try:
                git_p = parent_lookup(p)
            except KeyError:
                git_p = None
                if metadata is not None:
                    metadata.explicit_parent_ids = rev.parent_ids
            if git_p is not None:
                if len(git_p) != 40:
                    raise AssertionError("unexpected length for %r" % git_p)
                parents.append(git_p)
        commit.parents = parents
        try:
            encoding = rev.properties[u'git-explicit-encoding']
        except KeyError:
            encoding = rev.properties.get(u'git-implicit-encoding', 'utf-8')
        try:
            commit.encoding = rev.properties[u'git-explicit-encoding'].encode(
                'ascii')
        except KeyError:
            pass
        commit.committer = fix_person_identifier(rev.committer.encode(
            encoding))
        commit.author = fix_person_identifier(
            rev.get_apparent_authors()[0].encode(encoding))
        # TODO(jelmer): Don't use this hack.
        long = getattr(__builtins__, 'long', int)
        commit.commit_time = long(rev.timestamp)
        if u'author-timestamp' in rev.properties:
            commit.author_time = long(rev.properties[u'author-timestamp'])
        else:
            commit.author_time = commit.commit_time
        commit._commit_timezone_neg_utc = (
            u"commit-timezone-neg-utc" in rev.properties)
        commit.commit_timezone = rev.timezone
        commit._author_timezone_neg_utc = (
            u"author-timezone-neg-utc" in rev.properties)
        if u'author-timezone' in rev.properties:
            commit.author_timezone = int(rev.properties[u'author-timezone'])
        else:
            commit.author_timezone = commit.commit_timezone
        if u'git-gpg-signature' in rev.properties:
            commit.gpgsig = rev.properties[u'git-gpg-signature'].encode(
                'utf-8')
        if u'git-gpg-signature-b64' in rev.properties:
            commit.gpgsig = base64.b64decode(rev.properties[u'git-gpg-signature-b64'])
        commit.message = self._encode_commit_message(rev, rev.message,
                                                     encoding)
        if not isinstance(commit.message, bytes):
            raise TypeError(commit.message)
        if metadata is not None:
            try:
                mapping_registry.parse_revision_id(rev.revision_id)
            except errors.InvalidRevisionId:
                metadata.revision_id = rev.revision_id
            mapping_properties = set(
                [u'author', u'author-timezone', u'author-timezone-neg-utc',
                 u'commit-timezone-neg-utc', u'git-implicit-encoding',
                 u'git-gpg-signature', u'git-gpg-signature-b64',
                 u'git-explicit-encoding',
                 u'author-timestamp', u'file-modes'])
            for k, v in viewitems(rev.properties):
                if k not in mapping_properties:
                    metadata.properties[k] = v
        if not lossy and metadata:
            if self.roundtripping:
                commit.message = inject_bzr_metadata(commit.message, metadata,
                                                     encoding)
            else:
                raise NoPushSupport(
                    None, None, self, revision_id=rev.revision_id)
        if not isinstance(commit.message, bytes):
            raise TypeError(commit.message)
        i = 0
        propname = u'git-mergetag-0'
        while propname in rev.properties:
            commit.mergetag.append(Tag.from_string(rev.properties[propname]))
            i += 1
            propname = u'git-mergetag-%d' % i
        if u'git-extra' in rev.properties:
            commit.extra.extend(
                [l.split(b' ', 1)
                 for l in rev.properties[u'git-extra'].splitlines()])
        return commit

    def get_revision_id(self, commit):
        if commit.encoding:
            encoding = commit.encoding.decode('ascii')
        else:
            encoding = 'utf-8'
        try:
            message, metadata = self._decode_commit_message(
                None, commit.message, encoding)
        except UnicodeDecodeError:
            pass
        else:
            if metadata.revision_id:
                return metadata.revision_id
        return self.revision_id_foreign_to_bzr(commit.id)

    def import_commit(self, commit, lookup_parent_revid, strict=True):
        """Convert a git commit to a bzr revision.

        :return: a `breezy.revision.Revision` object, foreign revid and a
            testament sha1
        """
        if commit is None:
            raise AssertionError("Commit object can't be None")
        rev = ForeignRevision(commit.id, self,
                              self.revision_id_foreign_to_bzr(commit.id))
        rev.git_metadata = None

        def decode_using_encoding(rev, commit, encoding):
            rev.committer = commit.committer.decode(encoding)
            if commit.committer != commit.author:
                rev.properties[u'author'] = commit.author.decode(encoding)
            rev.message, rev.git_metadata = self._decode_commit_message(
                rev, commit.message, encoding)
        if commit.encoding is not None:
            rev.properties[u'git-explicit-encoding'] = commit.encoding.decode(
                'ascii')
            decode_using_encoding(rev, commit, commit.encoding.decode('ascii'))
        else:
            for encoding in ('utf-8', 'latin1'):
                try:
                    decode_using_encoding(rev, commit, encoding)
                except UnicodeDecodeError:
                    pass
                else:
                    if encoding != 'utf-8':
                        rev.properties[u'git-implicit-encoding'] = encoding
                    break
        if commit.commit_time != commit.author_time:
            rev.properties[u'author-timestamp'] = str(commit.author_time)
        if commit.commit_timezone != commit.author_timezone:
            rev.properties[u'author-timezone'] = "%d" % commit.author_timezone
        if commit._author_timezone_neg_utc:
            rev.properties[u'author-timezone-neg-utc'] = ""
        if commit._commit_timezone_neg_utc:
            rev.properties[u'commit-timezone-neg-utc'] = ""
        if commit.gpgsig:
            try:
                rev.properties[u'git-gpg-signature'] = commit.gpgsig.decode(
                    'utf-8')
            except UnicodeDecodeError:
                rev.properties[u'git-gpg-signature-b64'] = base64.b64encode(
                    commit.gpgsig)
        if commit.mergetag:
            for i, tag in enumerate(commit.mergetag):
                rev.properties[u'git-mergetag-%d' % i] = tag.as_raw_string()
        rev.timestamp = commit.commit_time
        rev.timezone = commit.commit_timezone
        rev.parent_ids = None
        if rev.git_metadata is not None:
            md = rev.git_metadata
            roundtrip_revid = md.revision_id
            if md.explicit_parent_ids:
                rev.parent_ids = md.explicit_parent_ids
            rev.properties.update(md.properties)
            verifiers = md.verifiers
        else:
            roundtrip_revid = None
            verifiers = {}
        if rev.parent_ids is None:
            parents = []
            for p in commit.parents:
                try:
                    parents.append(lookup_parent_revid(p))
                except KeyError:
                    parents.append(self.revision_id_foreign_to_bzr(p))
            rev.parent_ids = list(parents)
        unknown_extra_fields = []
        extra_lines = []
        for k, v in commit.extra:
            if k == HG_RENAME_SOURCE:
                extra_lines.append(k + b' ' + v + b'\n')
            elif k == HG_EXTRA:
                hgk, hgv = v.split(b':', 1)
                if hgk not in (HG_EXTRA_AMEND_SOURCE, ) and strict:
                    raise UnknownMercurialCommitExtra(commit, [hgk])
                extra_lines.append(k + b' ' + v + b'\n')
            else:
                unknown_extra_fields.append(k)
        if unknown_extra_fields and strict:
            raise UnknownCommitExtra(
                commit,
                [f.decode('ascii', 'replace') for f in unknown_extra_fields])
        if extra_lines:
            rev.properties[u'git-extra'] = b''.join(extra_lines)
        return rev, roundtrip_revid, verifiers


class BzrGitMappingv1(BzrGitMapping):
    revid_prefix = b'git-v1'
    experimental = False

    def __str__(self):
        return self.revid_prefix


class BzrGitMappingExperimental(BzrGitMappingv1):
    revid_prefix = b'git-experimental'
    experimental = True
    roundtripping = False

    BZR_DUMMY_FILE = '.bzrdummy'

    def _decode_commit_message(self, rev, message, encoding):
        if rev is None:
            rev = Revision()
        message = self._extract_hg_metadata(rev, message)
        message = self._extract_git_svn_metadata(rev, message)
        message, metadata = self._extract_bzr_metadata(rev, message)
        return message.decode(encoding), metadata

    def _encode_commit_message(self, rev, message, encoding):
        ret = message.encode(encoding)
        ret += self._generate_hg_message_tail(rev)
        ret += self._generate_git_svn_metadata(rev, encoding)
        return ret

    def import_commit(self, commit, lookup_parent_revid, strict=True):
        rev, roundtrip_revid, verifiers = super(
            BzrGitMappingExperimental, self).import_commit(
                commit, lookup_parent_revid, strict)
        rev.properties[u'converted_revision'] = "git %s\n" % commit.id
        return rev, roundtrip_revid, verifiers


class GitMappingRegistry(VcsMappingRegistry):
    """Registry with available git mappings."""

    def revision_id_bzr_to_foreign(self, bzr_revid):
        if bzr_revid == NULL_REVISION:
            from dulwich.protocol import ZERO_SHA
            return ZERO_SHA, None
        if not bzr_revid.startswith(b"git-"):
            raise errors.InvalidRevisionId(bzr_revid, None)
        (mapping_version, git_sha) = bzr_revid.split(b":", 1)
        mapping = self.get(mapping_version)
        return mapping.revision_id_bzr_to_foreign(bzr_revid)

    parse_revision_id = revision_id_bzr_to_foreign


mapping_registry = GitMappingRegistry()
mapping_registry.register_lazy(b'git-v1', __name__,
                               "BzrGitMappingv1")
mapping_registry.register_lazy(b'git-experimental',
                               __name__, "BzrGitMappingExperimental")
# Uncomment the next line to enable the experimental bzr-git mappings.
# This will make sure all bzr metadata is pushed into git, allowing for
# full roundtripping later.
# NOTE: THIS IS EXPERIMENTAL. IT MAY EAT YOUR DATA OR CORRUPT
# YOUR BZR OR GIT REPOSITORIES. USE WITH CARE.
# mapping_registry.set_default('git-experimental')
mapping_registry.set_default(b'git-v1')


class ForeignGit(ForeignVcs):
    """The Git Stupid Content Tracker"""

    @property
    def branch_format(self):
        from .branch import LocalGitBranchFormat
        return LocalGitBranchFormat()

    @property
    def repository_format(self):
        from .repository import GitRepositoryFormat
        return GitRepositoryFormat()

    def __init__(self):
        super(ForeignGit, self).__init__(mapping_registry)
        self.abbreviation = "git"

    @classmethod
    def serialize_foreign_revid(self, foreign_revid):
        return foreign_revid

    @classmethod
    def show_foreign_revid(cls, foreign_revid):
        return {"git commit": foreign_revid.decode('utf-8')}


foreign_vcs_git = ForeignGit()
default_mapping = mapping_registry.get_default()()


def symlink_to_blob(symlink_target):
    from dulwich.objects import Blob
    blob = Blob()
    if isinstance(symlink_target, text_type):
        symlink_target = symlink_target.encode('utf-8')
    blob.data = symlink_target
    return blob


def mode_is_executable(mode):
    """Check if mode should be considered executable."""
    return bool(mode & 0o111)


def mode_kind(mode):
    """Determine the Bazaar inventory kind based on Unix file mode."""
    if mode is None:
        return None
    entry_kind = (mode & 0o700000) / 0o100000
    if entry_kind == 0:
        return 'directory'
    elif entry_kind == 1:
        file_kind = (mode & 0o70000) / 0o10000
        if file_kind == 0:
            return 'file'
        elif file_kind == 2:
            return 'symlink'
        elif file_kind == 6:
            return 'tree-reference'
        else:
            raise AssertionError(
                "Unknown file kind %d, perms=%o." % (file_kind, mode,))
    else:
        raise AssertionError(
            "Unknown kind, perms=%r." % (mode,))


def object_mode(kind, executable):
    if kind == 'directory':
        return stat.S_IFDIR
    elif kind == 'symlink':
        mode = stat.S_IFLNK
        if executable:
            mode |= 0o111
        return mode
    elif kind == 'file':
        mode = stat.S_IFREG | 0o644
        if executable:
            mode |= 0o111
        return mode
    elif kind == 'tree-reference':
        from dulwich.objects import S_IFGITLINK
        return S_IFGITLINK
    else:
        raise AssertionError


def entry_mode(entry):
    """Determine the git file mode for an inventory entry."""
    return object_mode(entry.kind, getattr(entry, 'executable', False))


def extract_unusual_modes(rev):
    try:
        foreign_revid, mapping = mapping_registry.parse_revision_id(
            rev.revision_id)
    except errors.InvalidRevisionId:
        return {}
    else:
        return mapping.export_unusual_file_modes(rev)


def parse_git_svn_id(text):
    (head, uuid) = text.rsplit(" ", 1)
    (full_url, rev) = head.rsplit("@", 1)
    return (full_url, int(rev), uuid)


def needs_roundtripping(repo, revid):
    try:
        mapping_registry.parse_revision_id(revid)
    except errors.InvalidRevisionId:
        return True
    else:
        return False



# -*- coding: utf-8 -*-
#
# Breezy documentation build configuration file, created by
# sphinx-quickstart on Tue Jul 21 17:04:52 2009.
#
# This file is execfile()d with the current directory set to its containing dir.

import sys, os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path = [os.path.abspath('../..')] + sys.path

# Most of the configuration for Breezy docs is defined here ...
from breezy.doc_generate.conf import *

## Configuration specific to this site ##

# The locale code for this documentation set
brz_locale = 'en'

# Translations & supporting helper function
brz_titles = {
    u'Table of Contents (%s)': None,
    u'Breezy User Guide': None,
    u'Breezy User Reference': None,
    u'Breezy Release Notes': None,
    u'Breezy Upgrade Guide': None,
    u"Breezy System Administrator's Guide": None,
    u'Breezy in five minutes': None,
    u'Breezy Tutorial': None,
    u'Using Breezy With Launchpad': None,
    u'Centralized Workflow Tutorial': None,
    u"What's New in Breezy 2.1?": None,
    }
def brz_title(s):
    return brz_titles.get(s) or s

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = brz_title(u"Table of Contents (%s)") % (release,)

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {'index': 'index.html'}

# Output file base name for HTML help builder.
htmlhelp_basename = 'brz-%s' % (brz_locale,)

# Grouping the document tree into files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
brz_documents = [
    # Manuals
    ('user-guide/index', 'brz-%s-user-guide' % (brz_locale,),
     brz_title(u'Breezy User Guide'), brz_team, 'manual'),
    ('user-reference/index', 'brz-%s-user-reference' % (brz_locale,),
     brz_title(u'Breezy User Reference'), brz_team, 'manual'),
    ('release-notes/index', 'brz-%s-release-notes' % (brz_locale,),
     brz_title(u'Breezy Release Notes'), brz_team, 'manual'),
    ('upgrade-guide/index', 'brz-%s-upgrade-guide' % (brz_locale,),
     brz_title(u'Breezy Upgrade Guide'), brz_team, 'manual'),
    ('admin-guide/index', 'brz-%s-admin-guide' % (brz_locale,),
     brz_title(u"Breezy System Administrator's Guide"), brz_team, 'manual'),
    # Tutorials
    ('mini-tutorial/index', 'brz-%s-tutorial-mini' % (brz_locale,),
     brz_title(u'Breezy in five minutes'), brz_team, 'howto'),
    ('tutorials/tutorial', 'brz-%s-tutorial' % (brz_locale,),
     brz_title(u'Breezy Tutorial'), brz_team, 'howto'),
    ('tutorials/using_breezy_with_github',
     'brz-%s-tutorial-with-github' % (brz_locale,),
     brz_title(u'Using Breezy With GitHub'), brz_team, 'howto'),
    ('tutorials/using_breezy_with_launchpad',
     'brz-%s-tutorial-with-launchpad' % (brz_locale,),
     brz_title(u'Using Breezy With Launchpad'), brz_team, 'howto'),
    ('tutorials/centralized_workflow',
     'brz-%s-tutorial-centralized' % (brz_locale,),
     brz_title(u'Centralized Workflow Tutorial'), brz_team, 'howto'),
    ('whats-new/whats-new-in-3.0', 'brz-%s-whats-new' % (brz_locale,),
     brz_title(u"What's New in Breezy 3.0?"), brz_team, 'howto'),
]

latex_documents = [
    (start, target + '.tex', title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
    ]

texinfo_documents = [
    (start, target, title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
    ]

# List of documents that shouldn't be included in the build.
unused_docs = [
    # Subtopics that get included
    'upgrade-guide/overview',
    'upgrade-guide/data_migration',
    'upgrade-guide/tips_and_tricks',
    # Plain-style documentation generation stuff
    'release-notes/NEWS',
    'user-reference/brz_man',
    'user-guide/index-plain',
    'admin-guide/index-plain',
    # Templates
    'release-notes/release-template',
    'release-notes/series-template',
    # Miscellaneous
    'user-reference/readme',
]



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Some commands for debugging and repairing bzr repositories.

Some of these commands may be good candidates for adding to bzr itself, perhaps
as hidden commands.
"""

from __future__ import absolute_import

from ... import version_info  # noqa: F401
from ...commands import plugin_cmds


def test_suite():
    from unittest import TestSuite
    from .tests import test_suite
    result = TestSuite()
    result.addTest(test_suite())
    return result


plugin_cmds.register_lazy(
    'cmd_check_chk', [], __name__ + '.check_chk')
plugin_cmds.register_lazy(
    'cmd_chk_used_by', [], __name__ + '.chk_used_by')
plugin_cmds.register_lazy(
    'cmd_fetch_all_records', [], __name__ + '.fetch_all_records')
plugin_cmds.register_lazy(
    'cmd_file_refs', [], __name__ + '.file_refs')
plugin_cmds.register_lazy(
    'cmd_fix_missing_keys_for_stacking', [],
    __name__ + '.missing_keys_for_stacking_fixer')
plugin_cmds.register_lazy(
    'cmd_mirror_revs_into', [],
    __name__ + '.missing_keys_for_stacking_fixer')
plugin_cmds.register_lazy(
    'cmd_repo_has_key', [], __name__ + '.repo_has_key')
plugin_cmds.register_lazy(
    'cmd_repo_keys', [], __name__ + '.repo_keys')



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for bzr-git's object store."""

from __future__ import absolute_import

import os
import shutil
import stat

from dulwich.objects import (
    Blob,
    Tree,
    )

from ...branchbuilder import (
    BranchBuilder,
    )
from ...bzr.inventory import (
    InventoryDirectory,
    InventoryFile,
    )
from ...errors import (
    NoSuchRevision,
    )
from ...graph import (
    DictParentsProvider,
    Graph,
    )
from ...tests import (
    TestCase,
    TestCaseWithTransport,
    )
from ...tests.features import SymlinkFeature

from ..cache import (
    DictGitShaMap,
    )
from ..object_store import (
    BazaarObjectStore,
    LRUTreeCache,
    directory_to_tree,
    _check_expected_sha,
    _find_missing_bzr_revids,
    _tree_to_objects,
    )


class ExpectedShaTests(TestCase):

    def setUp(self):
        super(ExpectedShaTests, self).setUp()
        self.obj = Blob()
        self.obj.data = b"foo"

    def test_none(self):
        _check_expected_sha(None, self.obj)

    def test_hex(self):
        _check_expected_sha(
            self.obj.sha().hexdigest().encode('ascii'), self.obj)
        self.assertRaises(AssertionError, _check_expected_sha,
                          b"0" * 40, self.obj)

    def test_binary(self):
        _check_expected_sha(self.obj.sha().digest(), self.obj)
        self.assertRaises(AssertionError, _check_expected_sha,
                          b"x" * 20, self.obj)


class FindMissingBzrRevidsTests(TestCase):

    def _find_missing(self, ancestry, want, have):
        return _find_missing_bzr_revids(
            Graph(DictParentsProvider(ancestry)),
            set(want), set(have))

    def test_simple(self):
        self.assertEqual(set(), self._find_missing({}, [], []))

    def test_up_to_date(self):
        self.assertEqual(set(),
                         self._find_missing({"a": ["b"]}, ["a"], ["a"]))

    def test_one_missing(self):
        self.assertEqual(set(["a"]),
                         self._find_missing({"a": ["b"]}, ["a"], ["b"]))

    def test_two_missing(self):
        self.assertEqual(set(["a", "b"]),
                         self._find_missing({"a": ["b"], "b": ["c"]}, ["a"], ["c"]))

    def test_two_missing_history(self):
        self.assertEqual(set(["a", "b"]),
                         self._find_missing({"a": ["b"], "b": ["c"], "c": ["d"]},
                                            ["a"], ["c"]))


class LRUTreeCacheTests(TestCaseWithTransport):

    def setUp(self):
        super(LRUTreeCacheTests, self).setUp()
        self.branch = self.make_branch(".")
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        self.cache = LRUTreeCache(self.branch.repository)

    def test_get_not_present(self):
        self.assertRaises(NoSuchRevision, self.cache.revision_tree,
                          "unknown")

    def test_revision_trees(self):
        self.assertRaises(NoSuchRevision, self.cache.revision_trees,
                          ["unknown", "la"])

    def test_iter_revision_trees(self):
        self.assertRaises(NoSuchRevision, self.cache.iter_revision_trees,
                          ["unknown", "la"])

    def test_get(self):
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        revid = bb.build_snapshot(None,
                                  [('add', ('', None, 'directory', None)),
                                   ('add', ('foo', b'foo-id',
                                            'file', b'a\nb\nc\nd\ne\n')),
                                   ])
        bb.finish_series()
        tree = self.cache.revision_tree(revid)
        self.assertEqual(revid, tree.get_revision_id())


class BazaarObjectStoreTests(TestCaseWithTransport):

    def setUp(self):
        super(BazaarObjectStoreTests, self).setUp()
        self.branch = self.make_branch(".")
        self.store = BazaarObjectStore(self.branch.repository)

    def test_get_blob(self):
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        b = Blob()
        b.data = b'a\nb\nc\nd\ne\n'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        bb.build_snapshot(None,
                          [('add', ('', None, 'directory', None)),
                           ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                           ])
        bb.finish_series()
        # read locks cache
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        self.store.unlock()
        self.store.lock_read()
        self.assertEqual(b, self.store[b.id])

    def test_directory_converted_to_symlink(self):
        self.requireFeature(SymlinkFeature)
        b = Blob()
        b.data = b'trgt'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        tree = self.branch.controldir.create_workingtree()
        self.build_tree_contents([
            ('foo/', ),
            ('foo/bar', b'a\nb\nc\nd\ne\n')])
        tree.add(['foo', 'foo/bar'])
        revid1 = tree.commit('commit 1')
        shutil.rmtree('foo')
        os.symlink('trgt', 'foo')
        revid2 = tree.commit('commit 2')
        # read locks cache
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        self.store.unlock()
        self.store.lock_read()
        self.assertEqual(b, self.store[b.id])

    def test_get_raw(self):
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        b = Blob()
        b.data = b'a\nb\nc\nd\ne\n'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertRaises(KeyError, self.store.get_raw, b.id)
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        bb.build_snapshot(None,
                          [('add', ('', None, 'directory', None)),
                           ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                           ])
        bb.finish_series()
        # read locks cache
        self.assertRaises(KeyError, self.store.get_raw, b.id)
        self.store.unlock()
        self.store.lock_read()
        self.assertEqual(b.as_raw_string(), self.store.get_raw(b.id)[1])

    def test_contains(self):
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        b = Blob()
        b.data = b'a\nb\nc\nd\ne\n'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertFalse(b.id in self.store)
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        bb.build_snapshot(None,
                          [('add', ('', None, 'directory', None)),
                           ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                           ])
        bb.finish_series()
        # read locks cache
        self.assertFalse(b.id in self.store)
        self.store.unlock()
        self.store.lock_read()
        self.assertTrue(b.id in self.store)


class TreeToObjectsTests(TestCaseWithTransport):

    def setUp(self):
        super(TreeToObjectsTests, self).setUp()
        self.idmap = DictGitShaMap()

    def test_no_changes(self):
        tree = self.make_branch_and_tree('.')
        self.addCleanup(tree.lock_read().unlock)
        entries = list(_tree_to_objects(tree, [tree], self.idmap, {}))
        self.assertEqual([], entries)

    def test_with_gitdir(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['.git', 'foo'])
        tree.add(['.git', 'foo'])
        revid = tree.commit('commit')
        revtree = tree.branch.repository.revision_tree(revid)
        self.addCleanup(revtree.lock_read().unlock)
        entries = list(_tree_to_objects(revtree, [], self.idmap, {}))
        self.assertEqual(['foo', ''], [p[0] for p in entries])

    def test_merge(self):
        basis_tree = self.make_branch_and_tree('base')
        self.build_tree(['base/foo/'])
        basis_tree.add(['foo'])
        basis_rev = basis_tree.commit('foo')
        basis_revtree = basis_tree.branch.repository.revision_tree(basis_rev)

        tree_a = self.make_branch_and_tree('a')
        tree_a.pull(basis_tree.branch)

        self.build_tree(['a/foo/file1'])
        self.build_tree(['a/foo/subdir-a/'])
        os.symlink('.', 'a/foo/subdir-a/symlink')
        tree_a.add(['foo/subdir-a', 'foo/subdir-a/symlink'])

        tree_a.add(['foo/file1'])
        rev_a = tree_a.commit('commit a')
        revtree_a = tree_a.branch.repository.revision_tree(rev_a)

        with revtree_a.lock_read():
            entries = list(_tree_to_objects(revtree_a, [basis_revtree],
                           self.idmap, {}))
            objects = {path: obj for (path, obj, key) in entries}
            subdir_a = objects['foo/subdir-a']

        tree_b = self.make_branch_and_tree('b')
        tree_b.pull(basis_tree.branch)
        self.build_tree(['b/foo/subdir/'])
        os.symlink('.', 'b/foo/subdir/symlink')
        tree_b.add(['foo/subdir', 'foo/subdir/symlink'])
        rev_b = tree_b.commit('commit b')
        revtree_b = tree_b.branch.repository.revision_tree(rev_b)
        self.addCleanup(revtree_b.lock_read().unlock)

        with revtree_b.lock_read():
            list(_tree_to_objects(revtree_b, [basis_revtree], self.idmap, {}))

        with tree_a.lock_write():
            tree_a.merge_from_branch(tree_b.branch)
        rev_merge = tree_a.commit('merge')

        revtree_merge = tree_a.branch.basis_tree()
        self.addCleanup(revtree_merge.lock_read().unlock)

        entries = list(_tree_to_objects(
            revtree_merge,
            [tree_a.branch.repository.revision_tree(r)
             for r in revtree_merge.get_parent_ids()],
            self.idmap, {}))
        objects = {path: obj for (path, obj, key) in entries}
        self.assertEqual(set(['', 'foo', 'foo/subdir']), set(objects))
        self.assertEqual(
            (stat.S_IFDIR, subdir_a.id), objects['foo'][b'subdir-a'])


class DirectoryToTreeTests(TestCase):

    def test_empty(self):
        t = directory_to_tree('', [], None, {}, None, allow_empty=False)
        self.assertEqual(None, t)

    def test_empty_dir(self):
        child_ie = InventoryDirectory(b'bar', 'bar', b'bar')
        t = directory_to_tree('', [child_ie], lambda p, x: None, {}, None,
                              allow_empty=False)
        self.assertEqual(None, t)

    def test_empty_dir_dummy_files(self):
        child_ie = InventoryDirectory(b'bar', 'bar', b'bar')
        t = directory_to_tree('', [child_ie], lambda p, x: None, {}, ".mydummy",
                              allow_empty=False)
        self.assertTrue(".mydummy" in t)

    def test_empty_root(self):
        child_ie = InventoryDirectory(b'bar', 'bar', b'bar')
        t = directory_to_tree('', [child_ie], lambda p, x: None, {}, None,
                              allow_empty=True)
        self.assertEqual(Tree(), t)

    def test_with_file(self):
        child_ie = InventoryFile(b'bar', 'bar', b'bar')
        b = Blob.from_string(b"bla")
        t1 = directory_to_tree('', [child_ie], lambda p, x: b.id, {}, None,
                               allow_empty=False)
        t2 = Tree()
        t2.add(b"bar", 0o100644, b.id)
        self.assertEqual(t1, t2)

    def test_with_gitdir(self):
        child_ie = InventoryFile(b'bar', 'bar', b'bar')
        git_file_ie = InventoryFile(b'gitid', '.git', b'bar')
        b = Blob.from_string(b"bla")
        t1 = directory_to_tree('', [child_ie, git_file_ie],
                               lambda p, x: b.id, {}, None,
                               allow_empty=False)
        t2 = Tree()
        t2.add(b"bar", 0o100644, b.id)
        self.assertEqual(t1, t2)



# Copyright (C) 2006-2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""UI tests for brz ignore."""


import os
import re

from breezy import (
    ignores,
    osutils,
    )
import breezy
from breezy.branch import Branch
import breezy.bzr.bzrdir as bzrdir
from breezy.errors import BzrCommandError
from breezy.osutils import (
    pathjoin,
    )
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer
from breezy.tests import TestCaseWithTransport
from breezy.workingtree import WorkingTree


class TestCommands(TestCaseWithTransport):

    def test_ignore_absolutes(self):
        """'ignore' with an absolute path returns an error"""
        self.make_branch_and_tree('.')
        self.run_bzr_error(('brz: ERROR: NAME_PATTERN should not '
                            'be an absolute path\n',),
                           'ignore /crud')

    def test_ignore_directories(self):
        """ignoring a directory should ignore directory tree.

        Also check that trailing slashes on directories are stripped.
        """
        self.run_bzr('init')
        self.build_tree(['dir1/', 'dir1/foo',
                         'dir2/', 'dir2/bar',
                         'dir3/', 'dir3/baz'])
        self.run_bzr(['ignore', 'dir1', 'dir2/', 'dir4\\'])
        self.check_file_contents('.bzrignore', b'dir1\ndir2\ndir4\n')
        self.assertEqual(self.run_bzr('unknowns')[0], 'dir3\n')

    def test_ignore_patterns(self):
        tree = self.make_branch_and_tree('.')

        self.assertEqual(list(tree.unknowns()), [])

        # is_ignored() will now create the user global ignore file
        # if it doesn't exist, so make sure we ignore it in our tests
        ignores._set_user_ignores(['*.tmp'])

        self.build_tree_contents(
            [('foo.tmp', b'.tmp files are ignored by default')])
        self.assertEqual(list(tree.unknowns()), [])

        self.build_tree_contents([('foo.c', b'int main() {}')])
        self.assertEqual(list(tree.unknowns()), ['foo.c'])

        tree.add('foo.c')
        self.assertEqual(list(tree.unknowns()), [])

        # 'ignore' works when creating the .bzrignore file
        self.build_tree_contents([('foo.blah', b'blah')])
        self.assertEqual(list(tree.unknowns()), ['foo.blah'])
        self.run_bzr('ignore *.blah')
        self.assertEqual(list(tree.unknowns()), [])
        self.check_file_contents('.bzrignore', b'*.blah\n')

        # 'ignore' works when then .bzrignore file already exists
        self.build_tree_contents([('garh', b'garh')])
        self.assertEqual(list(tree.unknowns()), ['garh'])
        self.run_bzr('ignore garh')
        self.assertEqual(list(tree.unknowns()), [])
        self.check_file_contents('.bzrignore', b'*.blah\ngarh\n')

    def test_ignore_multiple_arguments(self):
        """'ignore' works with multiple arguments"""
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a', 'b', 'c', 'd'])
        self.assertEqual(list(tree.unknowns()), ['a', 'b', 'c', 'd'])
        self.run_bzr('ignore a b c')
        self.assertEqual(list(tree.unknowns()), ['d'])
        self.check_file_contents('.bzrignore', b'a\nb\nc\n')

    def test_ignore_no_arguments(self):
        """'ignore' with no arguments returns an error"""
        self.make_branch_and_tree('.')
        self.run_bzr_error(('brz: ERROR: ignore requires at least one '
                            'NAME_PATTERN or --default-rules.\n',),
                           'ignore')

    def test_ignore_default_rules(self):
        out, err = self.run_bzr(['ignore', '--default-rules'])
        reference_set = set(ignores.USER_DEFAULTS)
        output_set = set(out.rstrip().split('\n'))
        self.assertEqual(reference_set, output_set)
        self.assertEqual('', err)

    def test_ignore_versioned_file(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a', 'b'])
        tree.add('a')

        # test a single versioned file
        out, err = self.run_bzr('ignore a')
        self.assertEqual(out,
                         "Warning: the following files are version controlled"
                         " and match your ignore pattern:\na\n"
                         "These files will continue to be version controlled"
                         " unless you 'brz remove' them.\n")

        # test a single unversioned file
        out, err = self.run_bzr('ignore b')
        self.assertEqual(out, '')

        # test wildcards
        tree.add('b')
        out, err = self.run_bzr('ignore *')
        self.assertEqual(out,
                         "Warning: the following files are version controlled"
                         " and match your ignore pattern:\n.bzrignore\na\nb\n"
                         "These files will continue to be version controlled"
                         " unless you 'brz remove' them.\n")

    def test_ignored_versioned_file_matching_new_pattern(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a', 'b'])
        tree.add(['a', 'b'])
        self.run_bzr('ignore *')

        # If only the given pattern is used then only 'b' should match in
        # this case.
        out, err = self.run_bzr('ignore b')
        self.assertEqual(out,
                         "Warning: the following files are version controlled"
                         " and match your ignore pattern:\nb\n"
                         "These files will continue to be version controlled"
                         " unless you 'brz remove' them.\n")

    def test_ignore_directory(self):
        """Test --directory option"""
        tree = self.make_branch_and_tree('a')
        self.run_bzr(['ignore', '--directory=a', 'README'])
        self.check_file_contents('a/.bzrignore', b'README\n')

    def test_ignored_invalid_pattern(self):
        """Ensure graceful handling for invalid ignore pattern.

        Test case for #300062.
        Invalid pattern should show clear error message.
        Invalid pattern should not be added to .bzrignore file.
        """
        tree = self.make_branch_and_tree('.')
        out, err = self.run_bzr(['ignore', 'RE:*.cpp', 'foo', 'RE:['], 3)
        self.assertEqual(out, '')
        self.assertContainsRe(err,
                              r'Invalid ignore pattern.*RE:\*\.cpp.*RE:\[', re.DOTALL)
        self.assertNotContainsRe(err, 'foo', re.DOTALL)
        self.assertFalse(os.path.isfile('.bzrignore'))



# -*- coding: utf-8 -*-
#
# Breezy documentation build configuration file, created by
# sphinx-quickstart on Tue Jul 21 17:04:52 2009.
#
# This file is execfile()d with the current directory set to its containing dir.

import sys, os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path = [os.path.abspath('../..'), os.path.abspath('api')] + sys.path

# Most of the configuration for Breezy docs is defined here ...
from breezy.doc_generate.conf import *


## Configuration specific to this site ##

# The locale code for this documentation set
brz_locale = 'en'

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = u"Developer Document Catalog (%s)" % (release,)

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {'index': 'index.html'}

# Output file base name for HTML help builder.
htmlhelp_basename = 'brz-developers'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
brz_documents = [
    ('HACKING', 'brz-en-developer-guide', u'Breezy Developer Guide',
     u'Breezy Developers', 'manual'),
    ('testing', 'brz-en-testing-guide', u'Breezy Testing Guide',
     u'Breezy Developers', 'manual'),
    ('overview', 'brz-en-architecture-overview', u'Breezy Architecture Overview',
     u'Breezy Developers', 'howto'),
    ('integration', 'brz-en-integration-guide', u'Breezy Integration Guide',
     u'Breezy Developers', 'howto'),
]

latex_documents = [
    (start, target + '.tex', title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
    ]

texinfo_documents = [
    (start, target, title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
    ]

# List of documents that shouldn't be included in the build.
# Note: Maybe some of them *ought* to be linked in somewhere?
unused_docs = [
    'add',
    'annotate',
    'bundle-creation',
    'bundle-format4',
    'check',
    'commit',
    'diff',
    'directory-fingerprints',
    'gc',
    'index-plain',
    'incremental-push-pull',
    'initial-push-pull',
    'merge-scaling',
    'missing',
    'performance-roadmap-rationale',
    'performance-use-case-analysis',
    'planned-change-integration',
    'planned-performance-changes',
    'revert',
    'status',
    'uncommit',
]



# Copyright (C) 2007,2012 Canonical Ltd
# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""An adapter between a Git Branch and a Bazaar Branch"""

from __future__ import absolute_import

from io import BytesIO
from collections import defaultdict

from dulwich.config import (
    ConfigFile as GitConfigFile,
    parse_submodules,
    )

from dulwich.objects import (
    NotCommitError,
    ZERO_SHA,
    )
from dulwich.repo import check_ref_format

from .. import (
    branch,
    cleanup,
    config,
    controldir,
    errors,
    lock,
    repository as _mod_repository,
    revision,
    trace,
    transport,
    urlutils,
    )
from ..foreign import ForeignBranch
from ..revision import (
    NULL_REVISION,
    )
from ..sixish import (
    text_type,
    viewitems,
    )
from ..tag import (
    Tags,
    InterTags,
    )
from ..trace import (
    is_quiet,
    mutter,
    warning,
    )

from .config import (
    GitBranchConfig,
    GitBranchStack,
    )
from .errors import (
    NoPushSupport,
    )
from .push import (
    remote_divergence,
    )
from .refs import (
    branch_name_to_ref,
    is_tag,
    ref_to_branch_name,
    ref_to_tag_name,
    remote_refs_dict_to_tag_refs,
    tag_name_to_ref,
    )
from .unpeel_map import (
    UnpeelMap,
    )
from .urls import (
    git_url_to_bzr_url,
    bzr_url_to_git_url,
    )


def _calculate_revnos(branch):
    if branch._format.stores_revno():
        return True
    config = branch.get_config_stack()
    return config.get('calculate_revnos')


class GitPullResult(branch.PullResult):
    """Result of a pull from a Git branch."""

    def _lookup_revno(self, revid):
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        if not _calculate_revnos(self.target_branch):
            return None
        # Try in source branch first, it'll be faster
        with self.target_branch.lock_read():
            return self.target_branch.revision_id_to_revno(revid)

    @property
    def old_revno(self):
        return self._lookup_revno(self.old_revid)

    @property
    def new_revno(self):
        return self._lookup_revno(self.new_revid)


class InterTagsFromGitToRemoteGit(InterTags):

    @classmethod
    def is_compatible(klass, source, target):
        if not isinstance(source, GitTags):
            return False
        if not isinstance(target, GitTags):
            return False
        if getattr(target.branch.repository, "_git", None) is not None:
            return False
        return True

    def merge(self, overwrite=False, ignore_master=False, selector=None):
        if self.source.branch.repository.has_same_location(self.target.branch.repository):
            return {}, []
        updates = {}
        conflicts = []
        source_tag_refs = self.source.branch.get_tag_refs()

        def get_changed_refs(old_refs):
            ret = dict(old_refs)
            for ref_name, tag_name, peeled, unpeeled in (
                    source_tag_refs.iteritems()):
                if selector and not selector(tag_name):
                    continue
                if old_refs.get(ref_name) == unpeeled:
                    pass
                elif overwrite or ref_name not in old_refs:
                    ret[ref_name] = unpeeled
                    updates[tag_name] = self.target.branch.repository.lookup_foreign_revision_id(
                        peeled)
                    self.target.branch._tag_refs = None
                else:
                    conflicts.append(
                        (tag_name,
                         self.repository.lookup_foreign_revision_id(peeled),
                         self.target.branch.repository.lookup_foreign_revision_id(
                             old_refs[ref_name])))
            return ret
        self.target.branch.repository.controldir.send_pack(
            get_changed_refs, lambda have, want: [])
        return updates, set(conflicts)


class InterTagsFromGitToLocalGit(InterTags):

    @classmethod
    def is_compatible(klass, source, target):
        if not isinstance(source, GitTags):
            return False
        if not isinstance(target, GitTags):
            return False
        if getattr(target.branch.repository, "_git", None) is None:
            return False
        return True

    def merge(self, overwrite=False, ignore_master=False, selector=None):
        if self.source.branch.repository.has_same_location(self.target.branch.repository):
            return {}, []

        conflicts = []
        updates = {}
        source_tag_refs = self.source.branch.get_tag_refs()

        target_repo = self.target.branch.repository

        for ref_name, tag_name, peeled, unpeeled in source_tag_refs:
            if selector and not selector(tag_name):
                continue
            if target_repo._git.refs.get(ref_name) == unpeeled:
                pass
            elif overwrite or ref_name not in target_repo._git.refs:
                try:
                    updates[tag_name] = (
                        target_repo.lookup_foreign_revision_id(peeled))
                except KeyError:
                    trace.warning('%s does not point to a valid object',
                                  tag_name)
                    continue
                except NotCommitError:
                    trace.warning('%s points to a non-commit object',
                                  tag_name)
                    continue
                target_repo._git.refs[ref_name] = unpeeled or peeled
                self.target.branch._tag_refs = None
            else:
                try:
                    source_revid = self.source.branch.repository.lookup_foreign_revision_id(
                        peeled)
                    target_revid = target_repo.lookup_foreign_revision_id(
                        target_repo._git.refs[ref_name])
                except KeyError:
                    trace.warning('%s does not point to a valid object',
                                  ref_name)
                    continue
                except NotCommitError:
                    trace.warning('%s points to a non-commit object',
                                  tag_name)
                    continue
                conflicts.append((tag_name, source_revid, target_revid))
        return updates, set(conflicts)


class InterTagsFromGitToNonGit(InterTags):

    @classmethod
    def is_compatible(klass, source, target):
        if not isinstance(source, GitTags):
            return False
        if isinstance(target, GitTags):
            return False
        return True

    def merge(self, overwrite=False, ignore_master=False, selector=None):
        """See Tags.merge_to."""
        source_tag_refs = self.source.branch.get_tag_refs()
        if ignore_master:
            master = None
        else:
            master = self.target.branch.get_master_branch()
        with cleanup.ExitStack() as es:
            if master is not None:
                es.enter_context(master.lock_write())
            updates, conflicts = self._merge_to(
                self.target, source_tag_refs, overwrite=overwrite,
                selector=selector)
            if master is not None:
                extra_updates, extra_conflicts = self._merge_to(
                    master.tags, overwrite=overwrite,
                    source_tag_refs=source_tag_refs,
                    ignore_master=ignore_master, selector=selector)
                updates.update(extra_updates)
                conflicts.update(extra_conflicts)
            return updates, conflicts

    def _merge_to(self, to_tags, source_tag_refs, overwrite=False,
                  selector=None):
        unpeeled_map = defaultdict(set)
        conflicts = []
        updates = {}
        result = dict(to_tags.get_tag_dict())
        for ref_name, tag_name, peeled, unpeeled in source_tag_refs:
            if selector and not selector(tag_name):
                continue
            if unpeeled is not None:
                unpeeled_map[peeled].add(unpeeled)
            try:
                bzr_revid = self.source.branch.lookup_foreign_revision_id(peeled)
            except NotCommitError:
                continue
            if result.get(tag_name) == bzr_revid:
                pass
            elif tag_name not in result or overwrite:
                result[tag_name] = bzr_revid
                updates[tag_name] = bzr_revid
            else:
                conflicts.append((tag_name, bzr_revid, result[tag_name]))
        to_tags._set_tag_dict(result)
        if len(unpeeled_map) > 0:
            map_file = UnpeelMap.from_repository(to_tags.branch.repository)
            map_file.update(unpeeled_map)
            map_file.save_in_repository(to_tags.branch.repository)
        return updates, set(conflicts)


InterTags.register_optimiser(InterTagsFromGitToRemoteGit)
InterTags.register_optimiser(InterTagsFromGitToLocalGit)
InterTags.register_optimiser(InterTagsFromGitToNonGit)


class GitTags(Tags):
    """Ref-based tag dictionary."""

    def __init__(self, branch):
        self.branch = branch
        self.repository = branch.repository

    def get_tag_dict(self):
        ret = {}
        for (ref_name, tag_name, peeled, unpeeled) in (
                self.branch.get_tag_refs()):
            try:
                bzr_revid = self.branch.lookup_foreign_revision_id(peeled)
            except NotCommitError:
                continue
            else:
                ret[tag_name] = bzr_revid
        return ret

    def lookup_tag(self, tag_name):
        """Return the referent string of a tag"""
        # TODO(jelmer): Replace with something more efficient for local tags.
        td = self.get_tag_dict()
        try:
            return td[tag_name]
        except KeyError:
            raise errors.NoSuchTag(tag_name)


class LocalGitTagDict(GitTags):
    """Dictionary with tags in a local repository."""

    def __init__(self, branch):
        super(LocalGitTagDict, self).__init__(branch)
        self.refs = self.repository.controldir._git.refs

    def _set_tag_dict(self, to_dict):
        extra = set(self.refs.allkeys())
        for k, revid in viewitems(to_dict):
            name = tag_name_to_ref(k)
            if name in extra:
                extra.remove(name)
            try:
                self.set_tag(k, revid)
            except errors.GhostTagsNotSupported:
                pass
        for name in extra:
            if is_tag(name):
                del self.repository._git[name]

    def set_tag(self, name, revid):
        try:
            git_sha, mapping = self.branch.lookup_bzr_revision_id(revid)
        except errors.NoSuchRevision:
            raise errors.GhostTagsNotSupported(self)
        self.refs[tag_name_to_ref(name)] = git_sha
        self.branch._tag_refs = None

    def delete_tag(self, name):
        ref = tag_name_to_ref(name)
        if ref not in self.refs:
            raise errors.NoSuchTag(name)
        del self.refs[ref]
        self.branch._tag_refs = None


class GitBranchFormat(branch.BranchFormat):

    def network_name(self):
        return b"git"

    def supports_tags(self):
        return True

    def supports_leaving_lock(self):
        return False

    def supports_tags_referencing_ghosts(self):
        return False

    def tags_are_versioned(self):
        return False

    def get_foreign_tests_branch_factory(self):
        from .tests.test_branch import ForeignTestsBranchFactory
        return ForeignTestsBranchFactory()

    def make_tags(self, branch):
        try:
            return branch.tags
        except AttributeError:
            pass
        if getattr(branch.repository, "_git", None) is None:
            from .remote import RemoteGitTagDict
            return RemoteGitTagDict(branch)
        else:
            return LocalGitTagDict(branch)

    def initialize(self, a_controldir, name=None, repository=None,
                   append_revisions_only=None):
        raise NotImplementedError(self.initialize)

    def get_reference(self, controldir, name=None):
        return controldir.get_branch_reference(name=name)

    def set_reference(self, controldir, name, target):
        return controldir.set_branch_reference(target, name)

    def stores_revno(self):
        """True if this branch format store revision numbers."""
        return False

    supports_reference_locations = False


class LocalGitBranchFormat(GitBranchFormat):

    def get_format_description(self):
        return 'Local Git Branch'

    @property
    def _matchingcontroldir(self):
        from .dir import LocalGitControlDirFormat
        return LocalGitControlDirFormat()

    def initialize(self, a_controldir, name=None, repository=None,
                   append_revisions_only=None):
        from .dir import LocalGitDir
        if not isinstance(a_controldir, LocalGitDir):
            raise errors.IncompatibleFormat(self, a_controldir._format)
        return a_controldir.create_branch(
            repository=repository, name=name,
            append_revisions_only=append_revisions_only)


class GitBranch(ForeignBranch):
    """An adapter to git repositories for bzr Branch objects."""

    @property
    def control_transport(self):
        return self._control_transport

    @property
    def user_transport(self):
        return self._user_transport

    def __init__(self, controldir, repository, ref, format):
        self.repository = repository
        self._format = format
        self.controldir = controldir
        self._lock_mode = None
        self._lock_count = 0
        super(GitBranch, self).__init__(repository.get_mapping())
        self.ref = ref
        self._head = None
        self._user_transport = controldir.user_transport.clone('.')
        self._control_transport = controldir.control_transport.clone('.')
        self._tag_refs = None
        params = {}
        try:
            self.name = ref_to_branch_name(ref)
        except ValueError:
            self.name = None
            if self.ref is not None:
                params = {"ref": urlutils.escape(self.ref)}
        else:
            if self.name != "":
                params = {"branch": urlutils.escape(self.name)}
        for k, v in params.items():
            self._user_transport.set_segment_parameter(k, v)
            self._control_transport.set_segment_parameter(k, v)
        self.base = controldir.user_transport.base

    def _get_checkout_format(self, lightweight=False):
        """Return the most suitable metadir for a checkout of this branch.
        Weaves are used if this branch's repository uses weaves.
        """
        if lightweight:
            return controldir.format_registry.make_controldir("git")
        else:
            return controldir.format_registry.make_controldir("default")

    def get_child_submit_format(self):
        """Return the preferred format of submissions to this branch."""
        ret = self.get_config_stack().get("child_submit_format")
        if ret is not None:
            return ret
        return "git"

    def get_config(self):
        return GitBranchConfig(self)

    def get_config_stack(self):
        return GitBranchStack(self)

    def _get_nick(self, local=False, possible_master_transports=None):
        """Find the nick name for this branch.

        :return: Branch nick
        """
        if getattr(self.repository, '_git', None):
            cs = self.repository._git.get_config_stack()
            try:
                return cs.get((b"branch", self.name.encode('utf-8')),
                              b"nick").decode("utf-8")
            except KeyError:
                pass
        return self.name or u"HEAD"

    def _set_nick(self, nick):
        cf = self.repository._git.get_config()
        cf.set((b"branch", self.name.encode('utf-8')),
               b"nick", nick.encode("utf-8"))
        f = BytesIO()
        cf.write_to_file(f)
        self.repository._git._put_named_file('config', f.getvalue())

    nick = property(_get_nick, _set_nick)

    def __repr__(self):
        return "<%s(%r, %r)>" % (self.__class__.__name__, self.repository.base,
                                 self.name)

    def generate_revision_history(self, revid, last_rev=None,
                                  other_branch=None):
        if last_rev is not None:
            graph = self.repository.get_graph()
            if not graph.is_ancestor(last_rev, revid):
                # our previous tip is not merged into stop_revision
                raise errors.DivergedBranches(self, other_branch)

        self.set_last_revision(revid)

    def lock_write(self, token=None):
        if token is not None:
            raise errors.TokenLockingNotSupported(self)
        if self._lock_mode:
            if self._lock_mode == 'r':
                raise errors.ReadOnlyError(self)
            self._lock_count += 1
        else:
            self._lock_ref()
            self._lock_mode = 'w'
            self._lock_count = 1
        self.repository.lock_write()
        return lock.LogicalLockResult(self.unlock)

    def leave_lock_in_place(self):
        raise NotImplementedError(self.leave_lock_in_place)

    def dont_leave_lock_in_place(self):
        raise NotImplementedError(self.dont_leave_lock_in_place)

    def get_stacked_on_url(self):
        # Git doesn't do stacking (yet...)
        raise branch.UnstackableBranchFormat(self._format, self.base)

    def _get_push_origin(self, cs):
        """Get the name for the push origin.

        The exact behaviour is documented in the git-config(1) manpage.
        """
        try:
            return cs.get((b'branch', self.name.encode('utf-8')), b'pushRemote')
        except KeyError:
            try:
                return cs.get((b'branch', ), b'remote')
            except KeyError:
                try:
                    return cs.get((b'branch', self.name.encode('utf-8')), b'remote')
                except KeyError:
                    return b'origin'

    def _get_origin(self, cs):
        try:
            return cs.get((b'branch', self.name.encode('utf-8')), b'remote')
        except KeyError:
            return b'origin'

    def _get_related_push_branch(self, cs):
        remote = self._get_push_origin(cs)
        try:
            location = cs.get((b"remote", remote), b"url")
        except KeyError:
            return None

        return git_url_to_bzr_url(location.decode('utf-8'), ref=self.ref)

    def _get_related_merge_branch(self, cs):
        remote = self._get_origin(cs)
        try:
            location = cs.get((b"remote", remote), b"url")
        except KeyError:
            return None

        try:
            ref = cs.get((b"branch", remote), b"merge")
        except KeyError:
            ref = self.ref

        return git_url_to_bzr_url(location.decode('utf-8'), ref=ref)

    def _get_parent_location(self):
        """See Branch.get_parent()."""
        cs = self.repository._git.get_config_stack()
        return self._get_related_merge_branch(cs)

    def _write_git_config(self, cs):
        f = BytesIO()
        cs.write_to_file(f)
        self.repository._git._put_named_file('config', f.getvalue())

    def set_parent(self, location):
        cs = self.repository._git.get_config()
        remote = self._get_origin(cs)
        this_url = urlutils.strip_segment_parameters(self.user_url)
        target_url, branch, ref = bzr_url_to_git_url(location)
        location = urlutils.relative_url(this_url, target_url)
        cs.set((b"remote", remote), b"url", location)
        if branch:
            cs.set((b"branch", remote), b"merge", branch_name_to_ref(branch))
        elif ref:
            cs.set((b"branch", remote), b"merge", ref)
        else:
            # TODO(jelmer): Maybe unset rather than setting to HEAD?
            cs.set((b"branch", remote), b"merge", b'HEAD')
        self._write_git_config(cs)

    def break_lock(self):
        raise NotImplementedError(self.break_lock)

    def lock_read(self):
        if self._lock_mode:
            if self._lock_mode not in ('r', 'w'):
                raise ValueError(self._lock_mode)
            self._lock_count += 1
        else:
            self._lock_mode = 'r'
            self._lock_count = 1
        self.repository.lock_read()
        return lock.LogicalLockResult(self.unlock)

    def peek_lock_mode(self):
        return self._lock_mode

    def is_locked(self):
        return (self._lock_mode is not None)

    def _lock_ref(self):
        pass

    def _unlock_ref(self):
        pass

    def unlock(self):
        """See Branch.unlock()."""
        if self._lock_count == 0:
            raise errors.LockNotHeld(self)
        try:
            self._lock_count -= 1
            if self._lock_count == 0:
                if self._lock_mode == 'w':
                    self._unlock_ref()
                self._lock_mode = None
                self._clear_cached_state()
        finally:
            self.repository.unlock()

    def get_physical_lock_status(self):
        return False

    def last_revision(self):
        with self.lock_read():
            # perhaps should escape this ?
            if self.head is None:
                return revision.NULL_REVISION
            return self.lookup_foreign_revision_id(self.head)

    def _basic_push(self, target, overwrite=False, stop_revision=None,
                    tag_selector=None):
        return branch.InterBranch.get(self, target)._basic_push(
            overwrite, stop_revision, tag_selector=tag_selector)

    def lookup_foreign_revision_id(self, foreign_revid):
        try:
            return self.repository.lookup_foreign_revision_id(foreign_revid,
                                                              self.mapping)
        except KeyError:
            # Let's try..
            return self.mapping.revision_id_foreign_to_bzr(foreign_revid)

    def lookup_bzr_revision_id(self, revid):
        return self.repository.lookup_bzr_revision_id(
            revid, mapping=self.mapping)

    def get_unshelver(self, tree):
        raise errors.StoringUncommittedNotSupported(self)

    def _clear_cached_state(self):
        super(GitBranch, self)._clear_cached_state()
        self._tag_refs = None

    def _iter_tag_refs(self, refs):
        """Iterate over the tag refs.

        :param refs: Refs dictionary (name -> git sha1)
        :return: iterator over (ref_name, tag_name, peeled_sha1, unpeeled_sha1)
        """
        raise NotImplementedError(self._iter_tag_refs)

    def get_tag_refs(self):
        with self.lock_read():
            if self._tag_refs is None:
                self._tag_refs = list(self._iter_tag_refs())
            return self._tag_refs

    def import_last_revision_info_and_tags(self, source, revno, revid,
                                           lossy=False):
        """Set the last revision info, importing from another repo if necessary.

        This is used by the bound branch code to upload a revision to
        the master branch first before updating the tip of the local branch.
        Revisions referenced by source's tags are also transferred.

        :param source: Source branch to optionally fetch from
        :param revno: Revision number of the new tip
        :param revid: Revision id of the new tip
        :param lossy: Whether to discard metadata that can not be
            natively represented
        :return: Tuple with the new revision number and revision id
            (should only be different from the arguments when lossy=True)
        """
        push_result = source.push(
            self, stop_revision=revid, lossy=lossy, _stop_revno=revno)
        return (push_result.new_revno, push_result.new_revid)

    def reconcile(self, thorough=True):
        """Make sure the data stored in this branch is consistent."""
        from ..reconcile import ReconcileResult
        # Nothing to do here
        return ReconcileResult()


class LocalGitBranch(GitBranch):
    """A local Git branch."""

    def __init__(self, controldir, repository, ref):
        super(LocalGitBranch, self).__init__(controldir, repository, ref,
                                             LocalGitBranchFormat())

    def create_checkout(self, to_location, revision_id=None, lightweight=False,
                        accelerator_tree=None, hardlink=False):
        t = transport.get_transport(to_location)
        t.ensure_base()
        format = self._get_checkout_format(lightweight=lightweight)
        checkout = format.initialize_on_transport(t)
        if lightweight:
            from_branch = checkout.set_branch_reference(target_branch=self)
        else:
            policy = checkout.determine_repository_policy()
            policy.acquire_repository()
            checkout_branch = checkout.create_branch()
            checkout_branch.bind(self)
            checkout_branch.pull(self, stop_revision=revision_id)
            from_branch = None
        return checkout.create_workingtree(
            revision_id, from_branch=from_branch, hardlink=hardlink)

    def _lock_ref(self):
        self._ref_lock = self.repository._git.refs.lock_ref(self.ref)

    def _unlock_ref(self):
        self._ref_lock.unlock()

    def break_lock(self):
        self.repository._git.refs.unlock_ref(self.ref)

    def _gen_revision_history(self):
        if self.head is None:
            return []
        last_revid = self.last_revision()
        graph = self.repository.get_graph()
        try:
            ret = list(graph.iter_lefthand_ancestry(
                last_revid, (revision.NULL_REVISION, )))
        except errors.RevisionNotPresent as e:
            raise errors.GhostRevisionsHaveNoRevno(last_revid, e.revision_id)
        ret.reverse()
        return ret

    def _get_head(self):
        try:
            return self.repository._git.refs[self.ref]
        except KeyError:
            return None

    def _read_last_revision_info(self):
        last_revid = self.last_revision()
        graph = self.repository.get_graph()
        try:
            revno = graph.find_distance_to_null(
                last_revid, [(revision.NULL_REVISION, 0)])
        except errors.GhostRevisionsHaveNoRevno:
            revno = None
        return revno, last_revid

    def set_last_revision_info(self, revno, revision_id):
        self.set_last_revision(revision_id)
        self._last_revision_info_cache = revno, revision_id

    def set_last_revision(self, revid):
        if not revid or not isinstance(revid, bytes):
            raise errors.InvalidRevisionId(revision_id=revid, branch=self)
        if revid == NULL_REVISION:
            newhead = None
        else:
            (newhead, self.mapping) = self.repository.lookup_bzr_revision_id(
                revid)
            if self.mapping is None:
                raise AssertionError
        self._set_head(newhead)

    def _set_head(self, value):
        if value == ZERO_SHA:
            raise ValueError(value)
        self._head = value
        if value is None:
            del self.repository._git.refs[self.ref]
        else:
            self.repository._git.refs[self.ref] = self._head
        self._clear_cached_state()

    head = property(_get_head, _set_head)

    def get_push_location(self):
        """See Branch.get_push_location."""
        push_loc = self.get_config_stack().get('push_location')
        if push_loc is not None:
            return push_loc
        cs = self.repository._git.get_config_stack()
        return self._get_related_push_branch(cs)

    def set_push_location(self, location):
        """See Branch.set_push_location."""
        self.get_config().set_user_option('push_location', location,
                                          store=config.STORE_LOCATION)

    def supports_tags(self):
        return True

    def store_uncommitted(self, creator):
        raise errors.StoringUncommittedNotSupported(self)

    def _iter_tag_refs(self):
        """Iterate over the tag refs.

        :param refs: Refs dictionary (name -> git sha1)
        :return: iterator over (ref_name, tag_name, peeled_sha1, unpeeled_sha1)
        """
        refs = self.repository.controldir.get_refs_container()
        for ref_name, unpeeled in viewitems(refs.as_dict()):
            try:
                tag_name = ref_to_tag_name(ref_name)
            except (ValueError, UnicodeDecodeError):
                continue
            peeled = refs.get_peeled(ref_name)
            if peeled is None:
                peeled = unpeeled
            if not isinstance(tag_name, text_type):
                raise TypeError(tag_name)
            yield (ref_name, tag_name, peeled, unpeeled)

    def create_memorytree(self):
        from .memorytree import GitMemoryTree
        return GitMemoryTree(self, self.repository._git.object_store,
                             self.head)


def _quick_lookup_revno(local_branch, remote_branch, revid):
    if not isinstance(revid, bytes):
        raise TypeError(revid)
    # Try in source branch first, it'll be faster
    with local_branch.lock_read():
        if not _calculate_revnos(local_branch):
            return None
        try:
            return local_branch.revision_id_to_revno(revid)
        except errors.NoSuchRevision:
            graph = local_branch.repository.get_graph()
            try:
                return graph.find_distance_to_null(
                    revid, [(revision.NULL_REVISION, 0)])
            except errors.GhostRevisionsHaveNoRevno:
                if not _calculate_revnos(remote_branch):
                    return None
                # FIXME: Check using graph.find_distance_to_null() ?
                with remote_branch.lock_read():
                    return remote_branch.revision_id_to_revno(revid)


class GitBranchPullResult(branch.PullResult):

    def __init__(self):
        super(GitBranchPullResult, self).__init__()
        self.new_git_head = None
        self._old_revno = None
        self._new_revno = None

    def report(self, to_file):
        if not is_quiet():
            if self.old_revid == self.new_revid:
                to_file.write('No revisions to pull.\n')
            elif self.new_git_head is not None:
                to_file.write('Now on revision %d (git sha: %s).\n' %
                              (self.new_revno, self.new_git_head))
            else:
                to_file.write('Now on revision %d.\n' % (self.new_revno,))
        self._show_tag_conficts(to_file)

    def _lookup_revno(self, revid):
        return _quick_lookup_revno(self.target_branch, self.source_branch,
                                   revid)

    def _get_old_revno(self):
        if self._old_revno is not None:
            return self._old_revno
        return self._lookup_revno(self.old_revid)

    def _set_old_revno(self, revno):
        self._old_revno = revno

    old_revno = property(_get_old_revno, _set_old_revno)

    def _get_new_revno(self):
        if self._new_revno is not None:
            return self._new_revno
        return self._lookup_revno(self.new_revid)

    def _set_new_revno(self, revno):
        self._new_revno = revno

    new_revno = property(_get_new_revno, _set_new_revno)


class GitBranchPushResult(branch.BranchPushResult):

    def _lookup_revno(self, revid):
        return _quick_lookup_revno(self.source_branch, self.target_branch,
                                   revid)

    @property
    def old_revno(self):
        return self._lookup_revno(self.old_revid)

    @property
    def new_revno(self):
        new_original_revno = getattr(self, "new_original_revno", None)
        if new_original_revno:
            return new_original_revno
        if getattr(self, "new_original_revid", None) is not None:
            return self._lookup_revno(self.new_original_revid)
        return self._lookup_revno(self.new_revid)


class InterFromGitBranch(branch.GenericInterBranch):
    """InterBranch implementation that pulls from Git into bzr."""

    @staticmethod
    def _get_branch_formats_to_test():
        try:
            default_format = branch.format_registry.get_default()
        except AttributeError:
            default_format = branch.BranchFormat._default_format
        from .remote import RemoteGitBranchFormat
        return [
            (RemoteGitBranchFormat(), default_format),
            (LocalGitBranchFormat(), default_format)]

    @classmethod
    def _get_interrepo(self, source, target):
        return _mod_repository.InterRepository.get(
            source.repository, target.repository)

    @classmethod
    def is_compatible(cls, source, target):
        if not isinstance(source, GitBranch):
            return False
        if isinstance(target, GitBranch):
            # InterLocalGitRemoteGitBranch or InterToGitBranch should be used
            return False
        if (getattr(cls._get_interrepo(source, target), "fetch_objects", None)
                is None):
            # fetch_objects is necessary for this to work
            return False
        return True

    def fetch(self, stop_revision=None, fetch_tags=None, limit=None, lossy=False):
        self.fetch_objects(
            stop_revision, fetch_tags=fetch_tags, limit=limit, lossy=lossy)
        return _mod_repository.FetchResult()

    def fetch_objects(self, stop_revision, fetch_tags, limit=None, lossy=False, tag_selector=None):
        interrepo = self._get_interrepo(self.source, self.target)
        if fetch_tags is None:
            c = self.source.get_config_stack()
            fetch_tags = c.get('branch.fetch_tags')

        def determine_wants(heads):
            if stop_revision is None:
                try:
                    head = heads[self.source.ref]
                except KeyError:
                    self._last_revid = revision.NULL_REVISION
                else:
                    self._last_revid = self.source.lookup_foreign_revision_id(
                        head)
            else:
                self._last_revid = stop_revision
            real = interrepo.get_determine_wants_revids(
                [self._last_revid], include_tags=fetch_tags, tag_selector=tag_selector)
            return real(heads)
        pack_hint, head, refs = interrepo.fetch_objects(
            determine_wants, self.source.mapping, limit=limit,
            lossy=lossy)
        if (pack_hint is not None and
                self.target.repository._format.pack_compresses):
            self.target.repository.pack(hint=pack_hint)
        return head, refs

    def _update_revisions(self, stop_revision=None, overwrite=False, tag_selector=None):
        head, refs = self.fetch_objects(stop_revision, fetch_tags=None, tag_selector=tag_selector)
        if overwrite:
            prev_last_revid = None
        else:
            prev_last_revid = self.target.last_revision()
        self.target.generate_revision_history(
            self._last_revid, last_rev=prev_last_revid,
            other_branch=self.source)
        return head, refs

    def update_references(self, revid=None):
        if revid is None:
            revid = self.target.last_revision()
        tree = self.target.repository.revision_tree(revid)
        try:
            with tree.get_file('.gitmodules') as f:
                for path, url, section in parse_submodules(
                        GitConfigFile.from_file(f)):
                    self.target.set_reference_info(
                        tree.path2id(path.decode('utf-8')), url.decode('utf-8'),
                        path.decode('utf-8'))
        except errors.NoSuchFile:
            pass

    def _basic_pull(self, stop_revision, overwrite, run_hooks,
                    _override_hook_target, _hook_master, tag_selector=None):
        if overwrite is True:
            overwrite = set(["history", "tags"])
        elif not overwrite:
            overwrite = set()
        result = GitBranchPullResult()
        result.source_branch = self.source
        if _override_hook_target is None:
            result.target_branch = self.target
        else:
            result.target_branch = _override_hook_target
        with self.target.lock_write(), self.source.lock_read():
            # We assume that during 'pull' the target repository is closer than
            # the source one.
            (result.old_revno, result.old_revid) = \
                self.target.last_revision_info()
            result.new_git_head, remote_refs = self._update_revisions(
                stop_revision, overwrite=("history" in overwrite),
                tag_selector=tag_selector)
            tags_ret = self.source.tags.merge_to(
                self.target.tags, ("tags" in overwrite), ignore_master=True)
            if isinstance(tags_ret, tuple):
                result.tag_updates, result.tag_conflicts = tags_ret
            else:
                result.tag_conflicts = tags_ret
            (result.new_revno, result.new_revid) = \
                self.target.last_revision_info()
            self.update_references(revid=result.new_revid)
            if _hook_master:
                result.master_branch = _hook_master
                result.local_branch = result.target_branch
            else:
                result.master_branch = result.target_branch
                result.local_branch = None
            if run_hooks:
                for hook in branch.Branch.hooks['post_pull']:
                    hook(result)
            return result

    def pull(self, overwrite=False, stop_revision=None,
             possible_transports=None, _hook_master=None, run_hooks=True,
             _override_hook_target=None, local=False, tag_selector=None):
        """See Branch.pull.

        :param _hook_master: Private parameter - set the branch to
            be supplied as the master to pull hooks.
        :param run_hooks: Private parameter - if false, this branch
            is being called because it's the master of the primary branch,
            so it should not run its hooks.
        :param _override_hook_target: Private parameter - set the branch to be
            supplied as the target_branch to pull hooks.
        """
        # This type of branch can't be bound.
        bound_location = self.target.get_bound_location()
        if local and not bound_location:
            raise errors.LocalRequiresBoundBranch()
        source_is_master = False
        with cleanup.ExitStack() as es:
            es.enter_context(self.source.lock_read())
            if bound_location:
                # bound_location comes from a config file, some care has to be
                # taken to relate it to source.user_url
                normalized = urlutils.normalize_url(bound_location)
                try:
                    relpath = self.source.user_transport.relpath(normalized)
                    source_is_master = (relpath == '')
                except (errors.PathNotChild, urlutils.InvalidURL):
                    source_is_master = False
            if not local and bound_location and not source_is_master:
                # not pulling from master, so we need to update master.
                master_branch = self.target.get_master_branch(possible_transports)
                es.enter_context(master_branch.lock_write())
                # pull from source into master.
                master_branch.pull(self.source, overwrite, stop_revision,
                                   run_hooks=False)
            else:
                master_branch = None
            return self._basic_pull(stop_revision, overwrite, run_hooks,
                                    _override_hook_target,
                                    _hook_master=master_branch,
                                    tag_selector=tag_selector)

    def _basic_push(self, overwrite, stop_revision, tag_selector=None):
        if overwrite is True:
            overwrite = set(["history", "tags"])
        elif not overwrite:
            overwrite = set()
        result = branch.BranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.old_revno, result.old_revid = self.target.last_revision_info()
        result.new_git_head, remote_refs = self._update_revisions(
            stop_revision, overwrite=("history" in overwrite),
            tag_selector=tag_selector)
        tags_ret = self.source.tags.merge_to(
            self.target.tags, "tags" in overwrite, ignore_master=True,
            selector=tag_selector)
        (result.tag_updates, result.tag_conflicts) = tags_ret
        result.new_revno, result.new_revid = self.target.last_revision_info()
        self.update_references(revid=result.new_revid)
        return result


class InterGitBranch(branch.GenericInterBranch):
    """InterBranch implementation that pulls between Git branches."""

    def fetch(self, stop_revision=None, fetch_tags=None, limit=None, lossy=False):
        raise NotImplementedError(self.fetch)


class InterLocalGitRemoteGitBranch(InterGitBranch):
    """InterBranch that copies from a local to a remote git branch."""

    @staticmethod
    def _get_branch_formats_to_test():
        from .remote import RemoteGitBranchFormat
        return [
            (LocalGitBranchFormat(), RemoteGitBranchFormat())]

    @classmethod
    def is_compatible(self, source, target):
        from .remote import RemoteGitBranch
        return (isinstance(source, LocalGitBranch) and
                isinstance(target, RemoteGitBranch))

    def _basic_push(self, overwrite, stop_revision, tag_selector=None):
        result = GitBranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        if stop_revision is None:
            stop_revision = self.source.last_revision()

        def get_changed_refs(old_refs):
            old_ref = old_refs.get(self.target.ref, None)
            if old_ref is None:
                result.old_revid = revision.NULL_REVISION
            else:
                result.old_revid = self.target.lookup_foreign_revision_id(
                    old_ref)
            new_ref = self.source.repository.lookup_bzr_revision_id(
                stop_revision)[0]
            if not overwrite:
                if remote_divergence(
                        old_ref, new_ref,
                        self.source.repository._git.object_store):
                    raise errors.DivergedBranches(self.source, self.target)
            refs = {self.target.ref: new_ref}
            result.new_revid = stop_revision
            for name, sha in viewitems(
                    self.source.repository._git.refs.as_dict(b"refs/tags")):
                if tag_selector and not tag_selector(name):
                    continue
                if sha not in self.source.repository._git:
                    trace.mutter('Ignoring missing SHA: %s', sha)
                    continue
                refs[tag_name_to_ref(name)] = sha
            return refs
        self.target.repository.send_pack(
            get_changed_refs,
            self.source.repository._git.generate_pack_data)
        return result


class InterGitLocalGitBranch(InterGitBranch):
    """InterBranch that copies from a remote to a local git branch."""

    @staticmethod
    def _get_branch_formats_to_test():
        from .remote import RemoteGitBranchFormat
        return [
            (RemoteGitBranchFormat(), LocalGitBranchFormat()),
            (LocalGitBranchFormat(), LocalGitBranchFormat())]

    @classmethod
    def is_compatible(self, source, target):
        return (isinstance(source, GitBranch) and
                isinstance(target, LocalGitBranch))

    def fetch(self, stop_revision=None, fetch_tags=None, limit=None, lossy=False):
        interrepo = _mod_repository.InterRepository.get(
            self.source.repository, self.target.repository)
        if stop_revision is None:
            stop_revision = self.source.last_revision()
        if fetch_tags is None:
            c = self.source.get_config_stack()
            fetch_tags = c.get('branch.fetch_tags')
        determine_wants = interrepo.get_determine_wants_revids(
            [stop_revision], include_tags=fetch_tags)
        interrepo.fetch_objects(determine_wants, limit=limit, lossy=lossy)
        return _mod_repository.FetchResult()

    def _basic_push(self, overwrite=False, stop_revision=None, tag_selector=None):
        if overwrite is True:
            overwrite = set(["history", "tags"])
        elif not overwrite:
            overwrite = set()
        result = GitBranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.old_revid = self.target.last_revision()
        refs, stop_revision = self.update_refs(stop_revision)
        self.target.generate_revision_history(
            stop_revision,
            (result.old_revid if ("history" not in overwrite) else None),
            other_branch=self.source)
        tags_ret = self.source.tags.merge_to(
            self.target.tags,
            overwrite=("tags" in overwrite),
            selector=tag_selector)
        if isinstance(tags_ret, tuple):
            (result.tag_updates, result.tag_conflicts) = tags_ret
        else:
            result.tag_conflicts = tags_ret
        result.new_revid = self.target.last_revision()
        return result

    def update_refs(self, stop_revision=None):
        interrepo = _mod_repository.InterRepository.get(
            self.source.repository, self.target.repository)
        c = self.source.get_config_stack()
        fetch_tags = c.get('branch.fetch_tags')

        if stop_revision is None:
            result = interrepo.fetch(branches=[self.source.ref], include_tags=fetch_tags)
            try:
                head = result.refs[self.source.ref]
            except KeyError:
                stop_revision = revision.NULL_REVISION
            else:
                stop_revision = self.target.lookup_foreign_revision_id(head)
        else:
            result = interrepo.fetch(
                revision_id=stop_revision, include_tags=fetch_tags)
        return result.refs, stop_revision

    def pull(self, stop_revision=None, overwrite=False,
             possible_transports=None, run_hooks=True, local=False,
             tag_selector=None):
        # This type of branch can't be bound.
        if local:
            raise errors.LocalRequiresBoundBranch()
        if overwrite is True:
            overwrite = set(["history", "tags"])
        elif not overwrite:
            overwrite = set()

        result = GitPullResult()
        result.source_branch = self.source
        result.target_branch = self.target
        with self.target.lock_write(), self.source.lock_read():
            result.old_revid = self.target.last_revision()
            refs, stop_revision = self.update_refs(stop_revision)
            self.target.generate_revision_history(
                stop_revision,
                (result.old_revid if ("history" not in overwrite) else None),
                other_branch=self.source)
            tags_ret = self.source.tags.merge_to(
                self.target.tags, overwrite=("tags" in overwrite),
                selector=tag_selector)
            if isinstance(tags_ret, tuple):
                (result.tag_updates, result.tag_conflicts) = tags_ret
            else:
                result.tag_conflicts = tags_ret
            result.new_revid = self.target.last_revision()
            result.local_branch = None
            result.master_branch = result.target_branch
            if run_hooks:
                for hook in branch.Branch.hooks['post_pull']:
                    hook(result)
        return result


class InterToGitBranch(branch.GenericInterBranch):
    """InterBranch implementation that pulls into a Git branch."""

    def __init__(self, source, target):
        super(InterToGitBranch, self).__init__(source, target)
        self.interrepo = _mod_repository.InterRepository.get(source.repository,
                                                             target.repository)

    @staticmethod
    def _get_branch_formats_to_test():
        try:
            default_format = branch.format_registry.get_default()
        except AttributeError:
            default_format = branch.BranchFormat._default_format
        from .remote import RemoteGitBranchFormat
        return [
            (default_format, LocalGitBranchFormat()),
            (default_format, RemoteGitBranchFormat())]

    @classmethod
    def is_compatible(self, source, target):
        return (not isinstance(source, GitBranch) and
                isinstance(target, GitBranch))

    def _get_new_refs(self, stop_revision=None, fetch_tags=None,
                      stop_revno=None):
        if not self.source.is_locked():
            raise errors.ObjectNotLocked(self.source)
        if stop_revision is None:
            (stop_revno, stop_revision) = self.source.last_revision_info()
        elif stop_revno is None:
            try:
                stop_revno = self.source.revision_id_to_revno(stop_revision)
            except errors.NoSuchRevision:
                stop_revno = None
        if not isinstance(stop_revision, bytes):
            raise TypeError(stop_revision)
        main_ref = self.target.ref
        refs = {main_ref: (None, stop_revision)}
        if fetch_tags is None:
            c = self.source.get_config_stack()
            fetch_tags = c.get('branch.fetch_tags')
        for name, revid in viewitems(self.source.tags.get_tag_dict()):
            if self.source.repository.has_revision(revid):
                ref = tag_name_to_ref(name)
                if not check_ref_format(ref):
                    warning("skipping tag with invalid characters %s (%s)",
                            name, ref)
                    continue
                if fetch_tags:
                    # FIXME: Skip tags that are not in the ancestry
                    refs[ref] = (None, revid)
        return refs, main_ref, (stop_revno, stop_revision)

    def _update_refs(self, result, old_refs, new_refs, overwrite, tag_selector):
        mutter("updating refs. old refs: %r, new refs: %r",
               old_refs, new_refs)
        result.tag_updates = {}
        result.tag_conflicts = []
        ret = dict(old_refs)

        def ref_equals(refs, ref, git_sha, revid):
            try:
                value = refs[ref]
            except KeyError:
                return False
            if (value[0] is not None and
                git_sha is not None and
                    value[0] == git_sha):
                return True
            if (value[1] is not None and
                revid is not None and
                    value[1] == revid):
                return True
            # FIXME: If one side only has the git sha available and the other
            # only has the bzr revid, then this will cause us to show a tag as
            # updated that hasn't actually been updated.
            return False
        # FIXME: Check for diverged branches
        for ref, (git_sha, revid) in viewitems(new_refs):
            if ref_equals(ret, ref, git_sha, revid):
                # Already up to date
                if git_sha is None:
                    git_sha = old_refs[ref][0]
                if revid is None:
                    revid = old_refs[ref][1]
                ret[ref] = new_refs[ref] = (git_sha, revid)
            elif ref not in ret or overwrite:
                try:
                    tag_name = ref_to_tag_name(ref)
                except ValueError:
                    pass
                else:
                    if tag_selector and not tag_selector(tag_name):
                        continue
                    result.tag_updates[tag_name] = revid
                ret[ref] = (git_sha, revid)
            else:
                # FIXME: Check diverged
                diverged = False
                if diverged:
                    try:
                        name = ref_to_tag_name(ref)
                    except ValueError:
                        pass
                    else:
                        result.tag_conflicts.append(
                            (name, revid, ret[name][1]))
                else:
                    ret[ref] = (git_sha, revid)
        return ret

    def fetch(self, stop_revision=None, fetch_tags=None, lossy=False,
              limit=None):
        if stop_revision is None:
            stop_revision = self.source.last_revision()
        ret = []
        if fetch_tags:
            for k, v in viewitems(self.source.tags.get_tag_dict()):
                ret.append((None, v))
        ret.append((None, stop_revision))
        try:
            revidmap = self.interrepo.fetch_objects(ret, lossy=lossy, limit=limit)
        except NoPushSupport:
            raise errors.NoRoundtrippingSupport(self.source, self.target)
        return _mod_repository.FetchResult(revidmap={
            old_revid: new_revid
            for (old_revid, (new_sha, new_revid)) in revidmap.items()})

    def pull(self, overwrite=False, stop_revision=None, local=False,
             possible_transports=None, run_hooks=True, _stop_revno=None,
             tag_selector=None):
        result = GitBranchPullResult()
        result.source_branch = self.source
        result.target_branch = self.target
        with self.source.lock_read(), self.target.lock_write():
            new_refs, main_ref, stop_revinfo = self._get_new_refs(
                stop_revision, stop_revno=_stop_revno)

            def update_refs(old_refs):
                return self._update_refs(result, old_refs, new_refs, overwrite, tag_selector)
            try:
                result.revidmap, old_refs, new_refs = (
                    self.interrepo.fetch_refs(update_refs, lossy=False))
            except NoPushSupport:
                raise errors.NoRoundtrippingSupport(self.source, self.target)
            (old_sha1, result.old_revid) = old_refs.get(
                main_ref, (ZERO_SHA, NULL_REVISION))
            if result.old_revid is None:
                result.old_revid = self.target.lookup_foreign_revision_id(
                    old_sha1)
            result.new_revid = new_refs[main_ref][1]
            result.local_branch = None
            result.master_branch = self.target
            if run_hooks:
                for hook in branch.Branch.hooks['post_pull']:
                    hook(result)
        return result

    def push(self, overwrite=False, stop_revision=None, lossy=False,
             _override_hook_source_branch=None, _stop_revno=None,
             tag_selector=None):
        result = GitBranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.local_branch = None
        result.master_branch = result.target_branch
        with self.source.lock_read(), self.target.lock_write():
            new_refs, main_ref, stop_revinfo = self._get_new_refs(
                stop_revision, stop_revno=_stop_revno)

            def update_refs(old_refs):
                return self._update_refs(result, old_refs, new_refs, overwrite, tag_selector)
            try:
                result.revidmap, old_refs, new_refs = (
                    self.interrepo.fetch_refs(
                        update_refs, lossy=lossy, overwrite=overwrite))
            except NoPushSupport:
                raise errors.NoRoundtrippingSupport(self.source, self.target)
            (old_sha1, result.old_revid) = old_refs.get(
                main_ref, (ZERO_SHA, NULL_REVISION))
            if lossy or result.old_revid is None:
                result.old_revid = self.target.lookup_foreign_revision_id(
                    old_sha1)
            result.new_revid = new_refs[main_ref][1]
            (result.new_original_revno,
                result.new_original_revid) = stop_revinfo
            for hook in branch.Branch.hooks['post_push']:
                hook(result)
        return result


branch.InterBranch.register_optimiser(InterGitLocalGitBranch)
branch.InterBranch.register_optimiser(InterFromGitBranch)
branch.InterBranch.register_optimiser(InterToGitBranch)
branch.InterBranch.register_optimiser(InterLocalGitRemoteGitBranch)
